
/*
 * borg-classifier/db.c: database interaction source code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"
#if CONFIGURE_MYSQL == 1

/* the mysql connection. */
MYSQL *connection;

/* db_connect: connects to the mysql database. */
void db_connect (const char *database) {
  /* initialize the connection handle. */
  connection = mysql_init (NULL);
  if (!connection)
    return;

  /* connect to the server. */
  if (mysql_real_connect (connection,
                          MYSQL_HOSTNAME,
                          MYSQL_USERNAME,
                          MYSQL_PASSWORD,
                          database ? database : MYSQL_DATABASE,
                          0, NULL, 0) == NULL) {
    /* fail. */
    fprintf (stderr, "FAIL! unable to connect to the '%s' database!\n",
             database ? database : MYSQL_DATABASE);
    mysql_close (connection);
    connection = NULL;
    return;
  }
}

/* db_disconnect: disconnects from the mysql database. */
void db_disconnect (void) {
  /* just disconnect. */
  if (connection) mysql_close (connection);
  connection = NULL;
}

/* db_query: runs the mysql_query command on the current connection.
 * @format: the query/format string to use.
 * @...: printf-style list of variables required by the query string.
 */
int db_query (const char *format, ...) {
  /* declare required variables. */
  char *qstr;
  va_list vl;
  int res;

  /* make sure we're connected. */
  if (!connection) {
    /* fail. */
    fprintf (stderr, "FAIL! attempted to query before connecting!\n");
    exit (1);
  }

  /* init the va_list. */
  va_start (vl, format);

  /* allocate space for the query string.
   * NOTE: WE HAVE NO WAY OF KNOWING THE REQUIRED LENGTH OF THE QUERY STRING,
   *       SO WE SIMPLY USE A MAX LENGTH OPTION! THIS MAY CAUSE PROBLEMS DOWN
   *       THE ROAD, SO LOOK HERE IF A MALLOC() CORRUPTION ABORT OCCURS.
   */
  qstr = (char *) malloc (sizeof (char) * 8192);
  vsnprintf (qstr, 8192, format, vl);

  /* run the mysql query. */
  res = mysql_query (connection, qstr);

  /* free the query string. */
  free (qstr);

  /* deinit the va_list. */
  va_end (vl);

  /* return the exit status. */
  return res;
}

/* db_results_get: returns the results of a select or describe statement.
 */
struct db_results *db_results_get (void) {
  /* declare the variables. */
  unsigned long i, j, *lengths;
  struct db_results *results;
  MYSQL_RES *res;
  MYSQL_FIELD *fields;
  MYSQL_ROW row;

  /* allocate a new structure for results. */
  results = (struct db_results *) malloc (sizeof (struct db_results));

  /* set the default results values. */
  results->rows = 0;
  results->cols = 0;
  results->keys = NULL;
  results->values = NULL;

  /* store the result from the query. */
  res = mysql_store_result (connection);

  /* process the results set. */
  if (res) {
    /* get the number of rows and columns. */
    results->rows = mysql_num_rows (res);
    results->cols = mysql_num_fields (res);

    /* get the field metadata. */
    fields = mysql_fetch_fields (res);

    /* allocate space for the keys and value rows. */
    results->keys = (char **) malloc (sizeof (char *) * results->cols);
    results->values = (char ***) malloc (sizeof (char **) * results->rows);

    /* loop through the field names. */
    for (j = 0; j < results->cols; j++) {
      /* allocate memory for the key string. */
      results->keys[j] =
        (char *) malloc (sizeof (char) * fields[j].name_length + 1);

      /* copy the key string. */
      strcpy (results->keys[j], fields[j].name);
    }

    /* loop through the rows. */
    for (i = 0; i < results->rows; i++) {
      /* get the data for this row. */
      row = mysql_fetch_row (res);
      lengths = mysql_fetch_lengths (res);

      /* allocate the column for this row. */
      results->values[i] = (char **) malloc (sizeof (char *) * results->cols);

      /* loop through the columns. */
      for (j = 0; j < results->cols; j++) {
        /* allocate memory for the strings in the column. */
        results->values[i][j] =
          (char *) malloc (sizeof (char) * (lengths[j] + 1));

        /* copy the value string. */
        if (row[j])
          strcpy (results->values[i][j], row[j]);
        else
          results->values[i][j] = NULL;
      }
    }

    /* free the result from the query. */
    mysql_free_result (res);
  }

  /* return the structure. */
  return results;
}

/* db_results_free: frees the results from a mysql query.
 * @results: the results to free.
 */
void db_results_free (struct db_results *results) {
  /* declare loop variables. */
  unsigned long i, j;

  /* loop through the rows. */
  for (i = 0; i < results->rows; i++) {
    /* loop through the columns. */
    for (j = 0; j < results->cols; j++) {
      /* free the value. */
      free (results->values[i][j]);
    }

    /* free the value array. */
    free (results->values[i]);
  }

  /* loop again through the columns. */
  for (j = 0; j < results->cols; j++) {
    /* free the key array. */
    free (results->keys[j]);
  }

  /* free the allocated strings. */
  free (results->keys);
  free (results->values);

  /* free the container. */
  free (results);
}

/* db_results_field: returns the value of a given row's field.
 * @results: the results to use.
 * @row: the row number.
 * @field: the field name.
 */
char *db_results_field (struct db_results *results,
                        unsigned int row,
                        const char *field) {
  /* declare loop variable. */
  unsigned long j;

  /* umm, make sure we got a valid pointer. */
  if (!results) return NULL;

  /* see if the row index is valid. */
  if (row >= results->rows) return NULL;

  /* loop through the columns. */
  for (j = 0; j < results->cols; j++) {
    /* check the field name. */
    if (strcmp (results->keys[j], field) == 0) {
      /* return the given data value. */
      if (results->values[row][j])
        return results->values[row][j];
      else
        return NULL;
    }
  }

  /* nothing was found. */
  return NULL;
}

#endif
