
/*
 * borg-classifier/em.c: expectation-maximization clustering source code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* forward definitions. */
int em_initialization (struct em_inst *em, struct em_data *X);

/* em_expectation: the expectation step of the EM algorithm. */
double em_expectation (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long i, j, k;
  double prevLogL, sum, max, tmp;
  double **L;

  /* save the previous log-likelihood. */
  prevLogL = em->logL;

  /* loop through the clusters. */
  for (k = 0; k < em->K; k++) {
    /* decompose the covariance matrix. */
    L = stat_cholesky_decomp (em->sig[k], X->n);

    /* check the decomposition. */
    if (!L) {
      /* print a warning message and restart the algorithm. */
      return em_initialization (em, X);
    }

    /* calculate the log-determinant of the covariance matrix. */
    em->logdets[k] = stat_cholesky_logdet (L, X->n);

    /* loop through the number of points. */
    for (i = 0; i < X->m; i++) {
      /* loop through the dimensions, calculating the data vectors. */
      for (j = 0; j < X->n; j++)
        em->dv[j] = (X->d[i][j] - em->mu[k][j]);

      /* solve for the second intermediate vector. */
      em->dvv = stat_cholesky_solve (L, X->n, em->dv);

      /* loop through the dimensions, backsubstituting. */
      for (sum = 0.0, j = 0; j < X->n; j++)
        sum += pow (em->dvv[j], 2.0);

      /* calculate the responsibility matrix value. */
      em->p[i][k] = -0.5 * (sum + em->logdets[k]) + log (em->Pk[k]);

      /* free the second intermediate. */
      free (em->dvv);
    }

    /* free the decomposed matrix. */
    stat_cholesky_free (L, X->n);
  }

  /* normalize p and calculate the log-likelihood. */
  em->logL = 0.0;
  for (i = 0; i < X->m; i++) {
    /* initialize the max value. */
    max = -99.9e99;

    /* find the maximum responsibility. */
    for (k = 0; k < em->K; k++)
      if (em->p[i][k] > max) max = em->p[i][k];

    /* exponentate. */
    for (sum = 0.0, k = 0; k < em->K; k++) sum += exp (em->p[i][k] - max);

    /* save the tmp value. */
    tmp = max + log (sum);

    /* re-exponentate. */
    for (k = 0; k < em->K; k++) em->p[i][k] = exp (em->p[i][k] - tmp);
    em->logL += tmp;
  }

  /* store the difference in the log-likelihoods. */
  em->eps = em->logL - prevLogL;

  /* success. */
  return 0;
}

/* em_maximization: the maximization step of the EM algorithm. */
int em_maximization (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  signed long i, ja, jb, k;
  double weight, sum;

  /* loop over the clusters. */
  for (k = 0; k < em->K; k++) {
    /* calculate the un-normalized weight. */
    for (weight = 0.0, i = 0; i < X->m; i++)
      weight += em->p[i][k];

    /* calculate the new, normalized weight. */
    em->Pk[k] = weight / (double) X->m;

    /* loop through the dimensions. */
    for (ja = 0; ja < X->n; ja++) {
      /* calculate the un-normalized mean for the point. */
      for (sum = 0.0, i = 0; i < X->m; i++)
        sum += em->p[i][k] * X->d[i][ja];

      /* normalize the mean. */
      em->mu[k][ja] = sum / weight;

      /* loop again through the dimensions. */
      for (jb = 0; jb < X->n; jb++) {
        /* loop through the points. */
        for (sum = 0.0, i = 0; i < X->m; i++)
          sum += em->p[i][k] * (X->d[i][ja] - em->mu[k][ja])
                             * (X->d[i][jb] - em->mu[k][jb]);

        /* store the new covariance point. */
        em->sig[k][ja][jb] = sum / weight;
      }
    }
  }

  /* calculate the sum of the Pk values. (should be one) */
  for (sum = 0.0, k = 0; k < em->K; k++) {
    /* check the value. */
    if (em->Pk[k] <= EPSILON) {
      /* re-init. */
      return em_initialization (em, X);
    }

    /* add to the sum. */
    sum += em->Pk[k];
  }

  /* success. */
  return 0;
}

/* em_initialization: the initialization of the EM algorithm. */
int em_initialization (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long i, j, k;

  /* see if we are at a critical limit. */
  if (em->n_init > 100) {
    /* this is failure. */
    fprintf (stderr, "FAIL! reached maximum initialization count!\n");
    return -1;
  }

  /* print a nice message. */
  if (em->n_init <= 1)
    fprintf (stderr, "INFO: initializing algorithm parameters.\n");

  /* initialize our estimates for the clusters. */
  em->n_iter = 1;
  em->n_init++;
  for (k = 0; k < em->K; k++) {
    /* set a uniform prior over P(k). */
    em->Pk[k] = 1.0 / (double) em->K;

    /* loop through the covariance matrix rows... */
    for (i = 0; i < X->n; i++) {
      /* ... and columns. */
      for (j = 0; j < X->n; j++) em->sig[k][i][j] = 0.0;

      /* initialize just the diagonals. */
      em->sig[k][i][i] = 1.0e-3;

      /* randomly initialize the mean in this dimension. */
      em->mu[k][i] = (4.0 * rand_val ()) - 2.0;
    }
  }

  /* run the first EM iteration. */
  em->eps = 1.0e9;
  em->logL = -99.9e99;
  em_expectation (em, X);
  return em_maximization (em, X);
}

/* em_data_new: allocates a new em_data struct. */
struct em_data *em_data_new (void) {
  /* allocate a pointer to a new struct. */
  struct em_data *X = (struct em_data *) malloc (sizeof (struct em_data));
  if (!X) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to allocate memory for the data matrix.\n");
    return NULL;
  }

  /* read in the principle component analysis scores space data. */
  X->d = read_princomp (&(X->n));
  X->m = N;
  if (!X->d || !X->m || !X->n) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to read in pca results file!\n");
    return NULL;
  }

  /* return the new struct. */
  return X;
}

/* em_data_free: frees an allocated em_data struct. */
void em_data_free (struct em_data *X) {
  /* declare required variables. */
  unsigned long i;

  /* make sure we have a valid pointer. */
  if (!X) return;

  /* free the contained data. */
  for (i = 0; i < X->m; i++) free (X->d[i]);
  free (X->d);

  /* free the struct. */
  free (X);
}

/* em_inst_new: allocates a new em_inst struct. */
struct em_inst *em_inst_new (struct em_data *X) {
  /* declare required variables. */
  unsigned long i, k;
  struct em_inst *em;

  /* check the data struct. */
  if (!X) {
    /* fail out. */
    fprintf (stderr, "FAIL! handed a null data matrix pointer!\n");
    return NULL;
  }

  /* allocate a pointer to a new struct. */
  em = (struct em_inst *) malloc (sizeof (struct em_inst));
  if (!em) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to allocate memory for the instance.\n");
    return NULL;
  }

  /* check the conf keys. */
  conf_check (CONF_K);

  /* get the cluster count. */
  em->K = atol (conf_get (CONF_K));
  if (em->K < 1 || em->K > 100) {
    /* fail out. */
    fprintf (stderr, "FAIL! invalid cluster count of %lu passed!\n", em->K);
    return NULL;
  }

  /* initialize the iteration counters. */
  em->n_iter = 1;
  em->n_init = 0;

  /* allocate memory for the density means. */
  em->mu = (double **) malloc (sizeof (double *) * em->K);
  for (k = 0; k < em->K; k++)
    em->mu[k] = (double *) malloc (sizeof (double) * X->n);

  /* allocate memory for the density covariances. */
  em->sig = (double ***) malloc (sizeof (double **) * em->K);
  for (k = 0; k < em->K; k++) {
    /* allocate memory for the covariance row. */
    em->sig[k] = (double **) malloc (sizeof (double *) * X->n);
    for (i = 0; i < X->n; i++)
      em->sig[k][i] = (double *) malloc (sizeof (double) * X->n);
  }

  /* allocate memory for the log determinants. */
  em->logdets = (double *) malloc (sizeof (double) * em->K);

  /* allocate memory for the density weights. */
  em->Pk = (double *) malloc (sizeof (double) * em->K);

  /* allocate memory for the responsibility matrix. */
  em->p = (double **) malloc (sizeof (double *) * X->m);
  for (i = 0; i < X->m; i++)
    em->p[i] = (double *) malloc (sizeof (double) * em->K);

  /* allocate memory for the intermediate vector. */
  em->dv = (double *) malloc (sizeof (double) * X->n);

  /* return the new struct. */
  return em;
}

/* em_inst_free: frees an allocated em_inst struct. */
void em_inst_free (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long i, k;

  /* make sure we have a valid pointer. */
  if (!em || !X) return;

  /* free the density means. */
  for (k = 0; k < em->K; k++) free (em->mu[k]);
  free (em->mu);

  /* free the density covariances. */
  for (k = 0; k < em->K; k++) {
    /* free the covariance row. */
    for (i = 0; i < X->n; i++) free (em->sig[k][i]);
    free (em->sig[k]);
  }
  free (em->sig);

  /* free the log determinants. */
  free (em->logdets);

  /* free the density weights. */
  free (em->Pk);

  /* free the responsibility matrix. */
  for (i = 0; i < X->m; i++) free (em->p[i]);
  free (em->p);

  /* allocate memory for the intermediate vector. */
  free (em->dv);
}

/* em_execute: executes an expectation-maximization clustering. */
int em_execute (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long k;

  /* check the pointers. */
  if (!em || !X) {
    /* fail out. */
    fprintf (stderr, "FAIL! handed null em struct pointers!\n");
    return -1;
  }

  /* run the EM initialization. */
  if (em_initialization (em, X) < 0) {
    /* fail out. */
    fprintf (stderr, "WARN! first em initialization failed!\n");
    return -1;
  }

  /* run until we reach convergence. */
  while (em->eps > EPSILON &&
         (em->n_iter < 20 || em->n_init < 3) &&
         em->n_init < 100) {
    /* increment the iteration counter. */
    em->n_iter++;

    /* run the EM algo. */
    if (em_expectation (em, X) < 0) {
      /* fail out. */
      fprintf (stderr, "WARN! encountered failure from E-step!\n");
      return -1;
    }

    if (em_maximization (em, X) < 0) {
      /* fail out. */
      fprintf (stderr, "WARN! encountered failure from M-step!\n");
      return -1;
    }
  }

  /* check the final status. */
  if (em->eps > EPSILON) {
    /* fail out. */
    fprintf (stderr, "WARN! failed to reach convergence!\n");
    return -1;
  }

  /* output the log-likelihood. */
  fprintf (stderr, "INFO: maximized log-likelihood = %lf\n", em->logL);

  /* loop again to output the cluster weights. */
  fprintf (stderr, "INFO: maximized Pk = {");
  for (k = 0; k < em->K; k++)
    fprintf (stderr, "%lf%s", em->Pk[k], k == em->K - 1 ? "}\n" : ", ");

  /* success. */
  return 0;
}

/* em_dispersion: calculates the clustering dispersion. */
void em_dispersion (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long **idx, *n_idx;
  unsigned long i, j, k, ia, ib;
  double D, d;

  /* allocate memory for the index array. */
  idx = (unsigned long **) malloc (sizeof (unsigned long *) * em->K);
  n_idx = (unsigned long *) malloc (sizeof (unsigned long) * em->K);
  for (k = 0; k < em->K; k++) {
    /* allocate the cluster index array. */
    idx[k] = (unsigned long *) malloc (sizeof (unsigned long) * X->m);
    n_idx[k] = 0;
  }

  /* loop through the points. */
  for (i = 0; i < X->m; i++) {
    /* find the most responsible cluster. */
    for (k = 0, j = 1; j < em->K; j++)
      if (em->p[i][j] > em->p[i][k]) k = j;

    /* add this point into the index array. */
    idx[k][n_idx[k]] = i;
    n_idx[k]++;
  }

  /* loop through the clusters. */
  for (em->W = 0.0, k = 0; k < em->K; k++) {
    /* loop through the points. */
    for (D = 0.0, ia = 0; ia < n_idx[k]; ia++) {
      /* and again through the points. */
      for (ib = 0; ib < n_idx[k]; ib++) {
        /* don't uselessly compare the same point to itself. */
        if (ia == ib) continue;

        /* calculate the squared euclidean distance between the points. */
        for (d = 0.0, j = 0; j < X->n; j++)
          d += pow (X->d[idx[k][ia]][j] - X->d[idx[k][ib]][j], 2.0);

        /* add this to the total distance. */
        D += d;
      }
    }

    /* add the distance to the dispersion. */
    em->W += D / (2.0 * (double) n_idx[k]);
  }

  /* free the index array. */
  for (k = 0; k < em->K; k++) free (idx[k]);
  free (n_idx);
  free (idx);

  /* print the dispersion result. */
  fprintf (stderr, "INFO: clustering W = %lf\n", em->W);
}

/* em_print_clustering: outputs the clustering information. */
void em_print_clustering (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long i, j, k;
  FILE *fout;

  /* open the output file. */
  fout = fopen (conf_get (CONF_CLUSTER), "wb");
  if (!fout) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to open output file!\n");
    return;
  }

  /* loop through the points. */
  for (i = 0; i < X->m; i++) {
    /* find the most responsible cluster. */
    for (k = 0, j = 1; j < em->K; j++)
      if (em->p[i][j] > em->p[i][k]) k = j;

    /* print the heading. */
    fprintf (fout, "%lu,%lu", i, k);

    /* loop through the clusters. */
    for (k = 0; k < em->K; k++) {
      /* print the cluster responsibility. */
      fprintf (fout, ",%lf", em->p[i][k]);
    }

    /* newline. */
    fprintf (fout, "\n");
  }

  /* close the output file. */
  fclose (fout);
}

/* em_print_parameters: outputs the clustering parameters. */
void em_print_parameters (struct em_inst *em, struct em_data *X) {
  /* declare required variables. */
  unsigned long i, j, k;
  FILE *fout;

  /* open the output file. */
  fout = fopen (conf_get (CONF_PARMS), "wb");
  if (!fout) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to open output file!\n");
    return;
  }

  /* loop through the clusters. */
  for (k = 0; k < em->K; k++) {
    /* loop through the clusters again. */
    fprintf (fout, "%lf", em->mu[k][0]);
    for (j = 1; j < X->n; j++)
      fprintf (fout, ",%lf", em->mu[k][j]);

    /* loop through the covariance matrix. */
    fprintf (fout, " [");
    for (i = 0; i < X->n; i++) {
      /* print the entries in the row. */
      fprintf (fout, "%lf", em->sig[k][i][0]);
      for (j = 1; j < X->n; j++)
        fprintf (fout, " %lf", em->sig[k][i][j]);

      /* print the ending. */
      if (i == X->n - 1)
        fprintf (fout, "]\n");
      else
        fprintf (fout, "; ");
    }
  }

  /* close the output file. */
  fclose (fout);
}
