
/*
 * borg-classifier/gen-best-cluster.c: best cluster identification app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* the pca matrix. */
unsigned long Xm, Xn;
double **X;

/* the cluster index array. */
unsigned long *indices;

/* the mean and variance. */
double *W, mean, var;

/* main: application entry point. */
int main (int argc, char **argv) {
  /* declare required variables. */
  unsigned long i, nfile;
  double disp, mindisp;
  char buf[2048], *best;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_PRINCOMP);
  conf_check (CONF_CLUSTERS);

  /* read in the pca file. */
  X = read_princomp (&Xn);
  Xm = N;

  /* open the clusters file. */
  FILE *fcs = fopen (conf_get (CONF_CLUSTERS), "rb");

  /* init the variables. */
  disp = mindisp = 1.0e9;
  best = NULL;
  nfile = 0;

  /* init the stat variables. */
  W = (double *) malloc (sizeof (double));
  mean = 0.0;
  var = 0.0;

  /* loop through the file. */
  while (!feof (fcs)) {
    /* get the next line in the file. */
    if (fgets (buf, 2048, fcs)) {
      /* shave off the newline. */
      buf[strlen (buf) - 1] = '\0';

      /* read in the clustering file. */
      if (file_exists (buf)) {
        /* read in the clustering. */
        indices = read_clustering (buf);

        /* get the current clustering dispersion. */
        disp = stat_clustering_disp (X, Xm, Xn, indices);

        /* save the dispersion and resize the dispersion array. */
        W[nfile] = disp;
        W = (double *) realloc ((void *) W, sizeof (double) * (nfile + 2));

        /* print a nice message. */
        fprintf (stderr, "INFO: file %4lu has W = %lf (best = %lf)\n",
                 nfile + 1, disp, mindisp);

        /* see if we have a better dispersion. */
        if (disp < mindisp) {
          /* store the new dispersion. */
          if (best) free (best);
          best = str_dup (buf);
          mindisp = disp;
        }

        /* save the previous dispersion and free the clustering. */
        free (indices);
        nfile++;
      }
      else {
        /* print a warning. */
        fprintf (stderr, "WARN! file '%s' does not exist!\n", buf);
      }
    }
  }

  /* close the clusters file. */
  fclose (fcs);

  /* calculate the mean of the clustering. */
  for (i = 0; i < nfile; i++) mean += W[i];
  mean /= (double) nfile;

  /* calculate the standard deviation of the clustering. */
  for (i = 0; i < nfile; i++) var += pow (W[i] - mean, 2.0);
  var = sqrt (var / ((double) (nfile - 1)));

  /* print the best clustering file. */
  fprintf (stdout, "%s %lf %lf %lf\n", best, mindisp, mean, var);
  free (best);

  /* free the pca matrix. */
  for (i = 0; i < Xm; i++) free (X[i]);
  free (X);

  /* exit the program. */
  return 0;
}
