
/*
 * borg-classifier/gen-clusters.c: multi-serial clustering calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* the cluster mean and variance. */
unsigned long rmin;
double Wmin, Wmean, Wvar;

/* main: application entry point. */
int main (int argc, char **argv) {
  /* declare required variables. */
  struct em_data *X;
  struct em_inst **em;
  unsigned long R, r;
  FILE *fdisp;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_DISPERSION);
  conf_check (CONF_RUNS);

  /* get the number of runs. */
  R = atol (conf_get (CONF_RUNS));

  /* init the dispersion variables. */
  Wmin = 1.0e9;
  Wmean = 0.0;
  Wvar = 0.0;
  rmin = 0;

  /* allocate the data matrix structure. */
  X = em_data_new ();

  /* allocate the array of instances. */
  em = (struct em_inst **) malloc (sizeof (struct em_inst *) * R);

  /* loop through the runs. */
  for (r = 0; r < R; r++) {
    /* allocate the clustering instance. */
    em[r] = em_inst_new (X);

    /* print a kind message. */
    fprintf (stderr, "INFO: assigning %lu points to %lu clusters (%lu/%lu).\n",
             X->m, em[r]->K, r + 1, R);

    /* run the clustering algorithm. */
    while (em_execute (em[r], X) < 0) {
      /* rebuild the clustering instance. */
      rand_init ();
      em_inst_free (em[r], X);
      em[r] = em_inst_new (X);
    }

    /* calculate the dispersion of the result. */
    em_dispersion (em[r], X);
    if (em[r]->W < Wmin) {
      /* store the new minimal clustering. */
      Wmin = em[r]->W;
      rmin = r;
    }
  }

  /* print clustering results for the minimal clustering. */
  em_print_clustering (em[rmin], X);
  em_print_parameters (em[rmin], X);

  /* calculate the mean of the clustering. */
  for (r = 0; r < R; r++) Wmean += em[r]->W;
  Wmean /= (double) R;

  /* calculate the standard deviation of the clustering. */
  for (r = 0; r < R; r++) Wvar += pow (em[r]->W - Wmean, 2.0);
  Wvar = sqrt (Wvar / ((double) (R - 1)));

  /* print the dispersion. */
  fdisp = fopen (conf_get (CONF_DISPERSION), "wb");
  fprintf (fdisp, "%lf %lf %lf\n", Wmin, Wmean, Wvar);
  fclose (fdisp);

  /* free the instance array. */
  for (r = 0; r < R; r++) em_inst_free (em[r], X);
  free (em);

  /* free the data matrix. */
  em_data_free (X);

  /* exit the program. */
  return 0;
}
