
/*
 * borg-classifier/gen-consensus.c: consensus annotation calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

#define CONSENSUS_HEADER "\
<html>\n\
 <head>\n\
  <title>consensus results for K=%lu and C=%lu</title>\n\
  <style>\n\
body {\n\
 font-family: verdana, arial, helvetica, sans-serif;\n\
 font-weight: normal;\n\
 font-size: 12px;\n\
}\n\
.trWhite {\n\
 background-color: rgb(255,255,255);\n\
}\n\
.trGrey {\n\
 background-color: rgb(240,240,240);\n\
}\n\
  </style>\n\
 </head>\n\
 <body>\n\
  <table cellpadding=\"2\">\n\
"

#define CONSENSUS_ENTRY "\
   <tr class=\"%s\">\n\
     <td>\n\
       %.2lf%%\n\
     </td>\n\
     <td style=\"text-align: center;\">\n\
       (%lu/%lu)\n\
     </td>\n\
     <td>\n\
       %lu\n\
     </td>\n\
     <td>\n\
       %s\n\
     </td>\n\
     <td>\n\
       %s\n\
     </td>\n\
   </tr>\n\
"

#define CONSENSUS_FOOTER "\
  </table>\n\
 </body>\n\
</html>\n\
"

/* the number of clusters. */
unsigned long K;

/* the cluster index array. */
unsigned long *indices;
unsigned long **idx, *n_idx;

/* the counts along the dimensions. */
unsigned long *totals, *counts, *sorted, n_sorted;
double *percents;

/* heapsort_heapify: heapify the sorted heap. */
void heapsort_heapify (unsigned long k, unsigned long j) {
  /* setup our indices. */
  unsigned long left = (2 * j) + 1;
  unsigned long right = (2 * j) + 2;
  unsigned long max = j;
  signed long swap = -1;

  /* get the left-child score. */
  double sleft = left < n_sorted ? percents[sorted[left]] : -INFINITY;
  unsigned long cleft = left < n_sorted ? totals[sorted[left]] : -INFINITY;

  /* get the right-child score. */
  double sright = right < n_sorted ? percents[sorted[right]] : -INFINITY;
  unsigned long cright = left < n_sorted ? totals[sorted[right]] : -INFINITY;

  /* get the head score. */
  double smax = percents[sorted[max]];
  unsigned long cmax = totals[sorted[max]];

  /* see if the left child is larger than the parent. */
  if (left < n_sorted &&
      (sleft > smax || (sleft == smax && cleft > cmax))) {
    /* half-swap the nodes. */
    max = left;
    smax = sleft;
    cmax = cleft;
  }

  /* see if the right child is larger than either the left child or parent. */
  if (right < n_sorted &&
      (sright > smax || (sright == smax && cright > cmax))) {
    /* half-swap the nodes. */
    max = right;
    smax = sright;
    cmax = cright;
  }

  /* see if the largest node is not the root. */
  if (max != j) {
    /* swap the nodes. */
    swap = sorted[j];
    sorted[j] = sorted[max];
    sorted[max] = swap;

    /* max-heapify the child nodes. */
    heapsort_heapify (k, max);
  }
}

/* heapsort_get: extracts and deletes the sorted heap root node. */
signed long heapsort_get (unsigned long k) {
  /* save the heap index. */
  signed long root = sorted[0];

  /* move the last node to the root. */
  sorted[0] = sorted[n_sorted - 1];
  n_sorted--;

  /* re-heapify the search heap. */
  heapsort_heapify (k, 0);

  /* return the saved root index. */
  return root;
}

/* heapsort: sorts the dimension histogram. */
void heapsort (unsigned long k) {
  /* set the array indices up. */
  signed long j;
  for (j = 0; j < D; j++) sorted[j] = j;

  /* loop through the non-leaf nodes of the search heap, heapifying. */
  for (j = (signed long) floor ((double) n_sorted / 2.0) - 1; j >= 0; j--)
    heapsort_heapify (k, j);
}

/* main: application entry point. */
int main (int argc, char **argv) {
#if CONFIGURE_MYSQL == 1
  /* declare required variables. */
  unsigned long i, j, k, len;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_K);
  conf_check (CONF_CLUSTER);
  conf_check (CONF_CONSENSUS);

  /* read in the clustering file. */
  fprintf (stderr, "INFO: reading clustering from '%s'.\n",
           conf_get (CONF_CLUSTER));
  indices = read_clustering (conf_get (CONF_CLUSTER));

  /* allocate the counts array. */
  totals = (unsigned long *) malloc (sizeof (unsigned long) * D);
  counts = (unsigned long *) malloc (sizeof (unsigned long) * D);
  sorted = (unsigned long *) malloc (sizeof (unsigned long) * D);
  percents = (double *) malloc (sizeof (double) * D);
  n_sorted = D;

  /* loop through the index array to find the number of clusters. */
  for (K = 0, i = 0; i < N; i++) if (indices[i] > K) K = indices[i];
  K++;

  /* allocate some memory. */
  idx = (unsigned long **) malloc (sizeof (unsigned long *) * K);
  n_idx = (unsigned long *) malloc (sizeof (unsigned long) * K);
  for (k = 0; k < K; k++) {
    /* allocate the cluster indices. */
    idx[k] = (unsigned long *) malloc (sizeof (unsigned long) * N);
    n_idx[k] = 0;
  }

  /* loop through the points. */
  for (i = 0; i < N; i++) {
    /* store the cluster in its index. */
    k = indices[i];
    idx[k][n_idx[k]] = i;
    n_idx[k]++;
  }

  /* loop through the dimensions. */
  for (j = 0; j < D; j++) {
    /* and through the clusters. */
    for (totals[j] = 0, k = 0; k < K; k++) {
      /* and through the proteins in our cluster. */
      for (i = 0; i < n_idx[k]; i++) {
        /* increment the dimension if we have an annotation in this protein. */
        if (space[idx[k][i]][j]) totals[j]++;
      }
    }
  }

  /* get the number of clusters (k). */
  k = atol (conf_get (CONF_K)) - 1;

  /* loop through the dimenions. */
  for (j = 0; j < D; j++) {
    /* and through the proteins in our cluster. */
    for (counts[j] = 0, i = 0; i < n_idx[k]; i++) {
      /* increment the dimension if we have an annotation in this protein. */
      if (space[idx[k][i]][j]) counts[j]++;
    }

    /* divide out the count to make a percent. */
    percents[j] = (double) counts[j] * 100.0 / (double) totals[j];
  }

  /* sort the counts array and count the number of enriched dimensions. */
  heapsort (k);
  for (len = 0, j = 0; j < D; j++)
    if (percents[j] > (100.0 / (double) K)) len++;

  /* open the output file. */
  FILE *fout = fopen (conf_get (CONF_CONSENSUS), "wb");
  fprintf (fout, CONSENSUS_HEADER, K, k + 1);

  /* loop through the sorted dimensions. */
  j = 0;
  db_connect (NULL);
  char *description;
  unsigned long s;
  while (n_sorted) {
    /* get the next one. */
    s = heapsort_get (k);

    /* for enriched dimensions only... */
    if (percents[s] > (100.0 / (double) K)) {
      /* get the description and print the line. */
      description = dim_desc (s);
      fprintf (fout, CONSENSUS_ENTRY, j % 2 ? "trGrey" : "trWhite",
               percents[s], counts[s], totals[s], s, dims[s], description);

      /* free the description string. */
      free (description);
    }

    /* increment the counter. */
    j++;

    /* check the counter for periodic status updates. */
    if (!(j % 250) && j <= len)
      fprintf (stderr,
        "INFO: (cluster %lu/%lu) writing consensus data (%.0lf%% done)\n",
        k + 1, K, (double) j / (double) len * 100.0);
  }

  /* close the output file. */
  db_disconnect ();
  fprintf (fout, CONSENSUS_FOOTER);
  fclose (fout);

  /* free the counts array. */
  free (totals);
  free (counts);
  free (sorted);
  free (percents);

  /* free the clustering indices. */
  free (indices);
  for (k = 0; k < K; k++) free (idx[k]);
  free (n_idx);
  free (idx);

  /* exit the program. */
  return 0;
#else
  /* fail out. */
  fprintf (stderr, "FAIL! %s can't run without a mysql connection!\n", argv[0]);
  return 1;
#endif
}
