
/*
 * borg-classifier/gen-ellipse.c: 2*sigma cluster ellipse calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* constants for parsing the parms and calculating the ellipse vars. */
#define Nd 2
#define Z 2.0

/* constants for making points for the ellipses. */

/* the number of clusters. */
unsigned long K;

/* parsing variables. */
double t_mu[Nd];
double t_sig[Nd][Nd];

/* cluster and ellipse parms. */
gsl_vector *mu;
gsl_matrix *sig;
double phi;

/* main: application entry point. */
int main (int argc, char **argv) {
  /* declare required variables. */
  unsigned long i, j;
  double t, X, Y;
  char buf[2048];
  FILE *fparms;
  FILE *fout;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_PARMS);
  conf_check (CONF_ELLIPSE);

  /* print a nice message. */
  fprintf (stderr, "INFO: reading parameters from '%s'.\n",
           conf_get (CONF_PARMS));

  /* open the parms file. */
  fparms = fopen (conf_get (CONF_PARMS), "rb");
  if (!fparms) {
    /* fail. */
    fprintf (stderr, "FAIL! unable to open parms file!\n");
    return 1;
  }

  /* initialize the gsl variables. */
  mu = gsl_vector_calloc (Nd);
  sig = gsl_matrix_calloc (Nd, Nd);
  gsl_matrix *V = gsl_matrix_calloc (Nd, Nd);
  gsl_vector *S = gsl_vector_calloc (Nd);
  gsl_vector *work = gsl_vector_calloc (Nd);

  /* open the output file. */
  fout = fopen (conf_get (CONF_ELLIPSE), "wb");
  fprintf (fout, "\n# Surface 0 of 1 surfaces\n");
  fprintf (fout, "# Curve title: \"'e.dat'\"\n");

  /* loop through the lines of the parm file. */
  K = 0;
  while (!feof (fparms)) {
    /* get a line from the file. */
    if (fgets (buf, 2048, fparms)) {
      /* try to parse the file. */
      if (sscanf (buf,
          "%lf %lf %lf %lf %lf %lf\n",
          &t_mu[0], &t_mu[1],
          &t_sig[0][0], &t_sig[0][1],
          &t_sig[1][0], &t_sig[1][1]) == 6) {
        /* increment the cluster count. */
        K++;

        /* set up the gsl mean vector. */
        for (i = 0; i < Nd; i++) gsl_vector_set (mu, i, t_mu[i]);

        /* set up the gsl covariance matrix. */
        for (i = 0; i < Nd; i++)
          for (j = 0; j < Nd; j++)
            gsl_matrix_set (sig, i, j, t_sig[i][j]);

        /* decompose the covariance matrix. */
        gsl_linalg_SV_decomp (sig, V, S, work);

        /* calculate the elementwise square root of the S vector. */
        for (i = 0; i < Nd; i++)
          gsl_vector_set (S, i, Z * sqrt (gsl_vector_get (S, i)));

        /* calculate the rotation angle. */
        phi = acos (gsl_matrix_get (V, 0, 0));
        if (gsl_matrix_get (V, 1, 0) < 0.0) phi = 2.0 * M_PI - phi;
        /*phi = M_PI / 2.0 - phi;*/

        /* print a contour heading to the plot file. */
        fprintf (fout, "\n# Contour %lu, label: k%02lu\n", K - 1, K);

        /* loop through the t parameter. */
        for (t = 0; t <= 2 * M_PI; t += M_PI / 512) {
          /* calculate the X-Y pair. */
          X = gsl_vector_get (mu, 0)
            + gsl_vector_get (S, 0) * cos (t) * cos (phi)
            - gsl_vector_get (S, 1) * sin (t) * sin (phi);
          Y = gsl_vector_get (mu, 1)
            + gsl_vector_get (S, 0) * cos (t) * sin (phi)
            + gsl_vector_get (S, 1) * sin (t) * cos (phi);

          /* print the calculated pair. */
          fprintf (fout, " %lf  %lf  k%02lu\n", X, Y, K);
        }
      }
    }
  }

  /* close the files. */
  fclose (fparms);
  fclose (fout);

  /* clean up. */
  gsl_vector_free (mu);
  gsl_matrix_free (sig);

  /* exit the program. */
  return 0;
}
