
/*
 * borg-classifier/gen-entropy.c: known protein entropy calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* the cluster index array. */
unsigned long *indices;

/* the mean and variance. */
double *H;

/* main: application entry point. */
int main (int argc, char **argv) {
  /* declare required variables. */
  double H_max, H_mmp, H_top, H_egf;
  unsigned long c_mmp, c_top, c_egf;
  unsigned long i, j, k, K;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_CLUSTER);

  /* init the variables. */
  H_mmp = H_top = H_egf = 0.0;

  /* read in the clustering file. */
  if (file_exists (conf_get (CONF_CLUSTER))) {
    /* read in the clustering. */
    indices = read_clustering (conf_get (CONF_CLUSTER));

    /* loop through the proteins to find the cluster count. */
    for (K = 0, i = 0; i < N; i++) if (indices[i] > K) K = indices[i];
    K++;

    /* calculate the maximal entropy of the clustering. */
    H_max = log ((double) K) / log (2.0);

    /* loop through the clusters. */
    for (k = 0; k < K; k++) {
      /* zero the counts. */
      c_mmp = c_top = c_egf = 0;

      /* and through the points in each cluster. */
      for (i = 0; i < N; i++) {
        /* only in this cluster. */
        if (indices[i] != k) continue;

        /* count matrix metalloproteins. */
        for (j = 0; j < n_mmp; j++)
          if (mmp[j] == i) c_mmp++;

        /* count topoisomerases. */
        for (j = 0; j < n_top; j++)
          if (top[j] == i) c_top++;

        /* count endothelial growth factors. */
        for (j = 0; j < n_egf; j++)
          if (egf[j] == i) c_egf++;
      }

      /* add the entropy terms. */
      H_mmp -= (c_mmp ? ((double) c_mmp / (double) n_mmp)
                  * log ((double) c_mmp / (double) n_mmp) / log (2.0) : 0.0);
      H_top -= (c_top ? ((double) c_top / (double) n_top)
                  * log ((double) c_top / (double) n_top) / log (2.0) : 0.0);
      H_egf -= (c_egf ? ((double) c_egf / (double) n_egf)
                  * log ((double) c_egf / (double) n_egf) / log (2.0) : 0.0);
    }

    /* free the clustering. */
    free (indices);
  }
  else {
    /* print a failure and exit. */
    fprintf (stderr, "FAIL! file '%s' does not exist!\n",
             conf_get (CONF_CLUSTER));
    return 1;
  }

  /* normalize the entropies. */
  H_mmp /= (H_max ? H_max : 1.0);
  H_top /= (H_max ? H_max : 1.0);
  H_egf /= (H_max ? H_max : 1.0);

  /* print the entropies of the known proteins. */
  fprintf (stdout, "%lf %lf %lf\n", H_mmp, H_top, H_egf);

  /* exit the program. */
  return 0;
}
