
/*
 * borg-classifier/gen-pca-pls.c: nipals-based pca calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* declare an iteration counter. */
unsigned long n_iter;

/* the dimensionality of the principle component space. */
unsigned long L;

/* declare the required matrices and such. */
gsl_matrix *X, *E;
gsl_vector *t, *p;
gsl_vector *mu, *evar;
gsl_matrix *Evar;

/* main: application entry point. */
int main (int argc, char **argv) {
  /* declare required variables. */
  unsigned long i, j, l;
  char *scaling;
  double sum;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_PRINCOMP);
  conf_check (CONF_L);
  conf_check (CONF_SCALE);

  /* get the number of principle components (L). */
  L = atol (conf_get (CONF_L));
  scaling = conf_get (CONF_SCALE);

  /* see if we got a valid pc count. */
  if (L < 1 || L > D) {
    /* print an error and quit. */
    fprintf (stderr,
             "FAIL! invalid principle component count of %lu passed!\n", L);
    return 1;
  }

  /* initialize the space matrix. */
  fprintf (stderr, "INFO: initializing data matrix.\n");
  X = gsl_matrix_calloc (N, D);
  for (i = 0; i < N; i++)
    for (j = 0; j < D; j++)
      gsl_matrix_set (X, i, j, (double) space[i][j]);

  /* initialie the principle components matrix. */
  fprintf (stderr, "INFO: initializing error matrix.\n");
  E = gsl_matrix_calloc (N, D);

  /* initialize the mean vector. */
  fprintf (stderr, "INFO: initializing mean vector.\n");
  mu = gsl_vector_calloc (D);
  for (j = 0; j < D; j++) {
    /* loop through the points to calculate the mean. */
    for (sum = 0.0, i = 0; i < N; i++) sum += gsl_matrix_get (X, i, j);

    /* scale the mean and store the value. */
    sum /= (double) N;
    gsl_vector_set (mu, j, sum);

    /* subtract the mean from the data. */
    for (i = 0; i < N; i++)
      gsl_matrix_set (X, i, j, gsl_matrix_get (X, i, j) - sum);
  }

  /* should we do scaling? */
  if (strcmp (scaling, "none") == 0) {
    /* do no scaling. */
    fprintf (stderr, "INFO: not performing any data scaling.\n");
  }
  else if (strcmp (scaling, "var") == 0) {
    /* do variance scaling. */
    fprintf (stderr, "INFO: performing variance data scaling.\n");
    for (j = 0; j < D; j++) {
      /* loop through the points to calculate the variance. */
      for (sum = 0.0, i = 0; i < N; i++)
        sum += pow (gsl_matrix_get (X, i, j), 2.0);

      /* scale the (bias-corrected sample) variance. */
      sum /= (double) (N - 1);

      /* scale the data by the variance. */
      for (i = 0; i < N; i++)
        gsl_matrix_set (X, i, j, gsl_matrix_get (X, i, j) / sum);
    }
  }
  else if (strcmp (scaling, "stdev") == 0) {
    /* do variance scaling. */
    fprintf (stderr, "INFO: performing stdev data scaling.\n");
    for (j = 0; j < D; j++) {
      /* loop through the points to calculate the variance. */
      for (sum = 0.0, i = 0; i < N; i++)
        sum += pow (gsl_matrix_get (X, i, j), 2.0);

      /* scale the (bias-corrected sample) variance. */
      sum = sqrt (sum / (double) (N - 1));

      /* scale the data by the variance. */
      for (i = 0; i < N; i++)
        gsl_matrix_set (X, i, j, gsl_matrix_get (X, i, j) / sum);
    }
  }
  else if (strcmp (scaling, "pareto") == 0) {
    /* do variance scaling. */
    fprintf (stderr, "INFO: performing pareto data scaling.\n");
    for (j = 0; j < D; j++) {
      /* loop through the points to calculate the variance. */
      for (sum = 0.0, i = 0; i < N; i++)
        sum += pow (gsl_matrix_get (X, i, j), 2.0);

      /* scale the (bias-corrected sample) variance. */
      sum = sqrt (sqrt (sum / (double) (N - 1)));

      /* scale the data by the variance. */
      for (i = 0; i < N; i++)
        gsl_matrix_set (X, i, j, gsl_matrix_get (X, i, j) / sum);
    }
  }
  else {
    /* fail. */
    fprintf (stderr, "FAIL! invalid data scaling method '%s'\n", scaling);
    exit (1);
  }

  /* initialize the scores and loadings vectors. */
  fprintf (stderr, "INFO: initializing scores and loadings vectors.\n");
  t = gsl_vector_calloc (N);
  p = gsl_vector_calloc (D);
  evar = gsl_vector_calloc (L + 1);

  /* initialize the explained variances matrix. */
  fprintf (stderr, "INFO: initializing explained variances matrix.\n");
  Evar = gsl_matrix_calloc (L + 1, D);

  /* print a nice message. */
  fprintf (stderr, "INFO: finding the first %lu principle components.\n", L);

  /* initialize the scores and error. */
  gsl_matrix_get_col (t, X, 0);
  gsl_matrix_memcpy (E, X);

  /* declare a few more intermediates. */
  gsl_matrix *pc = gsl_matrix_calloc (N, D);
  gsl_matrix *S = gsl_matrix_calloc (N, L);
  gsl_matrix *P = gsl_matrix_calloc (L, D);
  double tdot, pdot, delta, tau, taupre;

  /* calculate the variance of the complete data. */
  for (j = 0; j < D; j++) {
    /* calculate over all the points. */
    for (sum = 0.0, i = 0; i < N; i++)
      sum += pow (gsl_matrix_get (E, i, j), 2.0);

    /* scale the sum to yeild a variance. */
    sum /= (double) (N - 1);

    /* store the variance. */
    gsl_matrix_set (Evar, 0, j, sum);
    gsl_vector_set (evar, 0, gsl_vector_get (evar, 0) + sum);
  }

  /* loop through the principle components. */
  for (l = 0; l < L; l++) {
    /* init the eigenvalues. */
    n_iter = 1;
    taupre = 100.0;
    tau = 100.0;
    delta = 0.0;

    /* randomly init the loadings. */
    for (i = 0; i < N; i++)
      gsl_vector_set (t, i, rand_val ());

    /* print a nice message. */
    fprintf (stderr, "INFO: iteratively calculating PC%lu.\n", l + 1);

    /* loop until we reach convergence. */
    while (n_iter < 2 || delta > EPSILON) {
      /* print a friendly message. */
      fprintf (stderr, "INFO: running for PC%lu, iteration %lu.",
               l + 1, n_iter);
      if (n_iter > 1)
        fprintf (stderr, " (delta = %lf)\n", delta);
      else
        fprintf (stderr, "\n");

      /* project X onto t to find the loading. p=(E'*t)/(t'*t) */
      gsl_blas_dgemv (CblasTrans, 1.0, E, t, 0.0, p);
      gsl_blas_ddot (t, t, &tdot);
      gsl_vector_scale (p, 1.0 / tdot);

      /* normalize the loading. p=p/sqrt(p'*p) */
      gsl_blas_ddot (p, p, &pdot);
      gsl_vector_scale (p, 1.0 / sqrt (pdot));

      /* project X onto p to find the score. t=(E*p)/(p'*p) */
      gsl_blas_dgemv (CblasNoTrans, 1.0, E, p, 0.0, t);
      gsl_blas_ddot (p, p, &pdot);
      gsl_vector_scale (t, 1.0 / pdot);

      /* get the new eigenvalue. */
      n_iter++;
      taupre = tau;
      gsl_blas_ddot (t, t, &tau);
      delta = fabs (tau - taupre);
    }

    /* save the scores. */
    for (i = 0; i < N; i++)
      gsl_matrix_set (S, i, l, gsl_vector_get (t, i));

    /* save the loadings. */
    for (j = 0; j < D; j++)
      gsl_matrix_set (P, l, j, gsl_vector_get (p, j));

    /* calculate the outer product of t, and p. (pc=t*p') */
    for (i = 0; i < N; i++)
      for (j = 0; j < D; j++)
        gsl_matrix_set (pc, i, j,
                        gsl_vector_get (t, i) *
                        gsl_vector_get (p, j));

    /* subtract this principle component from the data. */
    gsl_matrix_sub (E, pc);

    /* calculate the percent contributions. */
    for (j = 0; j < D; j++) {
      /* calculate squared residuals over all the points. */
      for (sum = 0.0, i = 0; i < N; i++)
        sum += pow (gsl_matrix_get (E, i, j), 2.0);

      /* scale the sum to yield a variance. */
      sum /= (double) (N - 1);

      /* store the variance. */
      gsl_matrix_set (Evar, l + 1, j, sum);
      gsl_vector_set (evar, l + 1, gsl_vector_get (evar, l + 1) + sum);
    }

    /* print the percent contributions. */
    fprintf (stderr, "INFO: PC%lu captures %.2lf%% of original variance.\n",
             l + 1,
             (gsl_vector_get (evar, l) - gsl_vector_get (evar, l + 1)) /
             gsl_vector_get (evar, 0) * 100.0);
  }

  /* print a nice message. */
  fprintf (stderr, "INFO: calculating contributions matrix elements.\n");

  /* allocate memory for the contributions vector. */
  gsl_vector *contrib;
  contrib = gsl_vector_calloc (D);

  /* loop through the dimensions. */
  for (j = 0; j < D; j++) {
    /* zero out the contribution. */
    gsl_vector_set (contrib, j, 0.0);

    /* calculate the actual variances due to each PC. */
    for (l = L; l > 0; l--)
      gsl_matrix_set (Evar, l, j,
        gsl_matrix_get (Evar, l - 1, j) - gsl_matrix_get (Evar, l, j));

    /* loop through the principle components. */
    for (sum = 0.0, l = 0; l < L; l++)
      sum += gsl_matrix_get (Evar, l + 1, j);

    /* store the summed value into the contribution vector. */
    gsl_vector_set (contrib, j, sum);
  }

  /* loop again through the dimensions. */
  for (sum = 0.0, j = 0; j < D; j++) sum += gsl_vector_get (contrib, j);

  /* normalize the contributions. */
  FILE *fcontrib = fopen (conf_get (CONF_CONTS), "wb");
  for (j = 0; j < D; j++) {
    /* calculate the normalized value. */
    gsl_vector_set (contrib, j, gsl_vector_get (contrib, j) / sum * 100.0);

    /* output the normalized contribution. */
    fprintf (fcontrib, "%lu,%lf\n", j, gsl_vector_get (contrib, j));
  }

  /* print a nice message. */
  fclose (fcontrib);
  fprintf (stderr, "INFO: orthonormalizing the loadings matrix.\n");

  /* declare elements for the QR decomp. */
  gsl_matrix *QR, *Q, *R, *Sorth;
  gsl_vector *vtau;
  gsl_matrix_view Porth;

  /* allocate memory for the QR decomposition elements. */
  Sorth = gsl_matrix_calloc (N, L);
  QR = gsl_matrix_calloc (D, L);
  Q = gsl_matrix_calloc (D, D);
  R = gsl_matrix_calloc (D, L);
  vtau = gsl_vector_calloc (L);

  /* calculate the QR decomp of the loadings and get the proper submatrix. */
  gsl_matrix_transpose_memcpy (QR, P);
  gsl_linalg_QR_decomp (QR, vtau);
  gsl_linalg_QR_unpack (QR, vtau, Q, R);
  Porth = gsl_matrix_submatrix (Q, 0, 0, D, L);

  /* transform the data using the orthonormalized loadings. */
  fprintf (stderr, "INFO: transforming data to scores space.\n");
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans,
                  1.0, X, &Porth.matrix,
                  0.0, Sorth);

  /* output the scores matrix. */
  fprintf (stderr, "INFO: writing scores space values.\n");
  FILE *fout = fopen (conf_get (CONF_PRINCOMP), "wb");
  for (i = 0; i < N; i++) {
    /* start the line. */
    fprintf (fout, "%.10lf", gsl_matrix_get (Sorth, i, 0));

    /* add the next columns to the line. */
    for (l = 1; l < L; l++)
      fprintf (fout, ",%.10lf", gsl_matrix_get (Sorth, i, l));

    /* finish the line. */
    fprintf (fout, "\n");
  }

  /* cleaning up. */
  fprintf (stderr, "INFO: cleaning up.\n");
  fclose (fout);
  gsl_matrix_free (X);
  gsl_matrix_free (E);
  gsl_matrix_free (Evar);
  gsl_vector_free (contrib);
  gsl_vector_free (mu);
  gsl_vector_free (t);
  gsl_vector_free (p);
  gsl_matrix_free (S);
  gsl_matrix_free (P);
  gsl_matrix_free (QR);
  gsl_matrix_free (Q);
  gsl_matrix_free (R);
  gsl_matrix_free (Sorth);

  /* exit the program. */
  return 0;
}
