
/*
 * borg-classifier/gen-pca-svd.c: svd-based pca calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* the dimensionality of the principle component space. */
unsigned long L;

/* the centered data matrix. */
gsl_matrix *X;

/* the mean. */
gsl_vector *mu;

/* the covariance matrix. */
gsl_matrix *sigma;

/* the results of the svd. */
gsl_matrix *V;
gsl_vector *S, *work;

/* the results of the pca. */
gsl_matrix *Y;

/* main: application entry point. */
int main (int argc, char **argv) {
  /* declare required variables. */
  unsigned long i, j, k, l;
  double sum;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* check the conf keys. */
  conf_check (CONF_PRINCOMP);
  conf_check (CONF_L);

  /* get the number of principle components (L). */
  L = atol (conf_get (CONF_L));

  /* see if we got a valid pc count. */
  if (L < 1 || L > D) {
    /* print an error and quit. */
    fprintf (stderr,
             "FAIL! invalid principle component count of %lu passed!\n", L);
    return 1;
  }

  /* initialize the mean vector. */
  fprintf (stderr, "INFO: initializing mean vector.\n");
  mu = gsl_vector_calloc (D);

  /* loop through the dimensions. */
  fprintf (stderr, "INFO: calculating mean vector.\n");
  for (j = 0; j < D; j++) {
    /* and through the proteins. */
    for (sum = 0.0, i = 0; i < N; i++) {
      /* add to our mean in this dimension. */
      sum += (double) space[i][j];
    }

    /* set the dimensional mean. */
    sum /= (double) N;
    gsl_vector_set (mu, j, sum);
  }

  /* initialize the data matrix. */
  fprintf (stderr, "INFO: initializing data matrix.\n");
  X = gsl_matrix_calloc (N, D);

  /* loop through the proteins. */
  for (i = 0; i < N; i++) {
    /* and through the dimensions. */
    for (j = 0; j < D; j++) {
      /* set the value of the data matrix. */
      sum = (double) space[i][j];
      sum -= gsl_vector_get (mu, j);
      gsl_matrix_set (X, i, j, sum);
    }
  }

  /* initialize the covariance matrix. */
  fprintf (stderr, "INFO: initializing covariance matrix.\n");
  sigma = gsl_matrix_calloc (D, D);

  /* loop through the rows of the covariance matrix. */
  fprintf (stderr, "INFO: calculating covariance matrix.\n");
  for (i = 0; i < D; i++) {
    /* loop through the columns of the covariance matrix. */
    for (j = 0; j < D; j++) {
      /* loop through the points. */
      for (sum = 0.0, k = 0; k < N; k++) {
        /* sum the products of the deviations up. */
        sum += ((double) space[k][i] - gsl_vector_get (mu, i)) *
               ((double) space[k][j] - gsl_vector_get (mu, j));
      }

      /* divide to make this a covariance. */
      sum /= (double) (N - 1);

      /* store the covariance in the gsl matrix. */
      gsl_matrix_set (sigma, i, j, sum);
    }
  }

  /* initialize the svd results. */
  fprintf (stderr, "INFO: initializing decomposition results.\n");
  V = gsl_matrix_calloc (D, D);
  S = gsl_vector_calloc (D);
  work = gsl_vector_calloc (D);

  /* calculate the decomposition. */
  fprintf (stderr, "INFO: calculating decomposition.\n");
  gsl_linalg_SV_decomp (sigma, V, S, work);

  /* calculate the percent variance captured. */
  gsl_vector *evar = gsl_vector_calloc (L + 1);
  for (sum = 0.0, j = 0; j < D; j++) sum += gsl_vector_get (S, j);
  gsl_vector_set (evar, 0, sum);
  for (l = 0; l < L; l++) {
    /* calculate the pc variance. */
    for (sum = 0.0, j = 0; j <= l; j++) sum += gsl_vector_get (S, j);
    gsl_vector_set (evar, l + 1, sum);

    /* output the explained variance. */
    fprintf (stderr, "INFO: PC%lu captures %.2lf%% of original variance.\n",
             l + 1,
             l == 0
             ? gsl_vector_get (evar, l + 1) / gsl_vector_get (evar, 0) * 100.0
             : (gsl_vector_get (evar, l + 1) - gsl_vector_get (evar, l)) /
               gsl_vector_get (evar, 0) * 100.0);
  }

  /* use the decomposition results. */
  fprintf (stderr, "INFO: transforming data to scores space.\n");
  Y = gsl_matrix_calloc (N, D);
  gsl_matrix_view Xv = gsl_matrix_submatrix (X, 0, 0, N - 1, D - 1);
  gsl_matrix_view Vv = gsl_matrix_submatrix (V, 0, 0, D - 1, D - 1);
  gsl_matrix_view Yv = gsl_matrix_submatrix (Y, 0, 0, N - 1, D - 1);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans,
                  1.0, &Xv.matrix, &Vv.matrix,
                  0.0, &Yv.matrix);

  /* loop through the output matrix. */
  fprintf (stderr, "INFO: outputting results.\n");
  FILE *fout = fopen (conf_get (CONF_PRINCOMP), "wb");
  for (i = 0; i < N; i++) {
    /* print the row, limiting dimensions. */
    fprintf (fout, "%.10lf", gsl_matrix_get (Y, i, 0));
    for (l = 1; l < L; l++)
      fprintf (fout, ",%.10lf", gsl_matrix_get (Y, i, l));

    /* print a newline. */
    fprintf (fout, "\n");
  }
  fclose (fout);

  /* free the initialized values. */
  fprintf (stderr, "INFO: cleaning up.\n");
  gsl_vector_free (mu);
  gsl_matrix_free (X);
  gsl_matrix_free (sigma);
  gsl_matrix_free (V);
  gsl_vector_free (S);
  gsl_vector_free (work);

  /* exit the program. */
  return 0;
}
