
/*
 * borg-classifier/gen-weights.c: protein weights calculation app code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

#define WEIGHTS_HEADER "\
<html>\n\
 <head>\n\
  <title>weighting results for K=%lu and C=%lu</title>\n\
  <style>\n\
body {\n\
 font-family: verdana, arial, helvetica, sans-serif;\n\
 font-weight: normal;\n\
 font-size: 12px;\n\
}\n\
.trWhite {\n\
 background-color: rgb(255,255,255);\n\
}\n\
.trGrey {\n\
 background-color: rgb(240,240,240);\n\
}\n\
  </style>\n\
 </head>\n\
 <body>\n\
  <table cellpadding=\"2\">\n\
"

#define WEIGHTS_ENTRY "\
   <tr class=\"%s\">\n\
     <td style=\"width: 200px;\">\n\
       <a href=\"http://www.uniprot.org/uniprot/%s\">%s</a>\n\
     </td>\n\
     <td style=\"width: 200px;%s\">\n\
       %.4lf\n\
     </td>\n\
     <td style=\"width: 200px;%s\">\n\
       %.4lf\n\
     </td>\n\
   </tr>\n\
"

#define WEIGHTS_FOOTER "\
  </table>\n\
 </body>\n\
</html>\n\
"

/* the number of clusters. */
unsigned long K;

/* the cluster index array. */
unsigned long *indices;
unsigned long **idx, *n_idx;

/* the weights we calculated. */
unsigned long *sorted, n_sorted;
double *lengths, *Ws, *Wd;

/* heapsort_heapify: heapify the sorted heap. */
void heapsort_heapify (unsigned long k, unsigned long j) {
  /* setup our indices. */
  unsigned long left = (2 * j) + 1;
  unsigned long right = (2 * j) + 2;
  unsigned long max = j;
  signed long swap = -1;

  /* get the left-child score. */
  double dleft = left < n_sorted ? Wd[sorted[left]] : -INFINITY;
  double sleft = left < n_sorted ? Ws[sorted[left]] : -INFINITY;

  /* get the right-child score. */
  double dright = right < n_sorted ? Wd[sorted[right]] : -INFINITY;
  double sright = right < n_sorted ? Ws[sorted[right]] : -INFINITY;

  /* get the head score. */
  double dmax = Wd[sorted[max]];
  double smax = Ws[sorted[max]];

  /* see if the left child is larger than the parent. */
  if (left < n_sorted &&
      (dleft > dmax || (dleft == dmax && sleft > smax))) {
    /* half-swap the nodes. */
    max = left;
    dmax = dleft;
    smax = sleft;
  }

  /* see if the right child is larger than either the left child or parent. */
  if (right < n_sorted &&
      (dright > dmax || (dright == dmax && sright > smax))) {
    /* half-swap the nodes. */
    max = right;
    dmax = dright;
    smax = sright;
  }

  /* see if the largest node is not the root. */
  if (max != j) {
    /* swap the nodes. */
    swap = sorted[j];
    sorted[j] = sorted[max];
    sorted[max] = swap;

    /* max-heapify the child nodes. */
    heapsort_heapify (k, max);
  }
}

/* heapsort_get: extracts and deletes the sorted heap root node. */
signed long heapsort_get (unsigned long k) {
  /* save the heap index. */
  signed long root = sorted[0];

  /* move the last node to the root. */
  sorted[0] = sorted[n_sorted - 1];
  n_sorted--;

  /* re-heapify the search heap. */
  heapsort_heapify (k, 0);

  /* return the saved root index. */
  return root;
}

/* heapsort: sorts the dimension histogram. */
void heapsort (unsigned long k) {
  /* loop through the non-leaf nodes of the search heap, heapifying. */
  signed long j;
  for (j = (signed long) floor ((double) n_sorted / 2.0) - 1; j >= 0; j--)
    heapsort_heapify (k, j);
}

/* main: application entry point. */
int main (int argc, char **argv) {
#if CONFIGURE_MYSQL == 1
  /* declare required variables. */
  unsigned long i, j, k;
  double max, sdev, ddev;

  /* initialize the core structures. */
  util_init_core (argc, argv);

  /* init the crosslink code. */
  xlink_init ();

  /* check the conf keys. */
  conf_check (CONF_K);
  conf_check (CONF_CLUSTER);
  conf_check (CONF_WEIGHTS);

  /* read in the clustering file. */
  fprintf (stderr, "INFO: reading clustering from '%s'.\n",
           conf_get (CONF_CLUSTER));
  indices = read_clustering (conf_get (CONF_CLUSTER));

  /* loop through the index array to find the number of clusters. */
  for (K = 0, i = 0; i < N; i++) if (indices[i] > K) K = indices[i];
  K++;

  /* allocate some memory. */
  idx = (unsigned long **) malloc (sizeof (unsigned long *) * K);
  n_idx = (unsigned long *) malloc (sizeof (unsigned long) * K);
  for (k = 0; k < K; k++) {
    /* allocate the cluster indices. */
    idx[k] = (unsigned long *) malloc (sizeof (unsigned long) * N);
    n_idx[k] = 0;
  }

  /* connect to the database. */
  db_connect (NULL);

  /* loop through the points. */
  for (i = 0; i < N; i++) {
    /* only include non-pdb structured proteins. */
    if (!uid_has_pdb (i)) {
      /* store the cluster in its index. */
      k = indices[i];
      idx[k][n_idx[k]] = i;
      n_idx[k]++;
    }
  }

  /* disconnect from the database. */
  db_disconnect ();

  /* get the number of clusters (k). */
  k = atol (conf_get (CONF_K)) - 1;

  /* allocate memory for the weights. */
  lengths = (double *) malloc (sizeof (double) * N);
  Ws = (double *) malloc (sizeof (double) * n_idx[k]);
  Wd = (double *) malloc (sizeof (double) * n_idx[k]);

  /* generate the sorted index array. */
  sorted = (unsigned long *) malloc (sizeof (unsigned long) * n_idx[k]);
  for (i = 0; i < n_idx[k]; i++) sorted[i] = i;
  n_sorted = n_idx[k];

  /* loop through the proteins. */
  fprintf (stderr, "INFO: calculating protein distances from origin.\n");
  for (max = 0.0, i = 0; i < N; i++) {
    /* calculate the length. */
    for (lengths[i] = 0.0, j = 0; j < D; j++)
      if (space[i][j]) lengths[i]++;

    /* see if we found a greater length. */
    if (lengths[i] > max) max = lengths[i];
  }

  /* normalize the lengths. */
  for (i = 0; i < N; i++) lengths[i] /= max;

  /* find the string interactions. */
  unsigned long *link, n;
  fprintf (stderr, "INFO: calculating protein weights.\n");
  for (i = 0; i < n_idx[k]; i++) {
    /* zero the weights. */
    Ws[i] = 0.0;
    Wd[i] = 0.0;

    /* get the string interactions. */
    link = xlink_get_string (idx[k][i], &n);
    if (link) {
      /* loop through the proteins in the link array. */
      for (j = 0; j < n; j++)
        Ws[i] += lengths[link[j]];

      /* free the link array. */
      free (link);
    }

    /* get the dip interactions. */
    link = xlink_get_dip (idx[k][i], &n);
    if (link) {
      /* loop through the proteins in the link array. */
      for (j = 0; j < n; j++)
        Wd[i] += lengths[link[j]];

      /* free the link array. */
      free (link);
    }

    /* divide the weights by the length of the protein. */
    Ws[i] /= lengths[i];
    Wd[i] /= lengths[i];

    /* see if we should give some stimulus. */
    if (i && !(i % 100))
      fprintf (stderr,
        "INFO: (cluster %lu/%lu) calculating protein weights (%.2lf%% done)\n",
        k + 1, K, (double) i / (double) n_idx[k] * 100.0);
  }

  /* calculate the mean of the weights array. */
  for (sdev = 0.0, i = 0; i < n_idx[k]; i++) sdev += Ws[i];
  for (ddev = 0.0, i = 0; i < n_idx[k]; i++) ddev += Wd[i];
  sdev /= (double) n_idx[k];
  ddev /= (double) n_idx[k];

  /* sort the weights array. */
  heapsort (k);

  /* open the output file. */
  FILE *fout = fopen (conf_get (CONF_WEIGHTS), "wb");
  fprintf (fout, WEIGHTS_HEADER, K, k + 1);

  /* loop through the sorted weights. */
  unsigned long dim_tm = dim_index ("keyw:TRANSMEMBRANE");
  unsigned long dim_3ds = dim_index ("keyw:3D-STRUCTURE");
  unsigned long s;
  j = 0;
  while (n_sorted) {
    /* get the next one. */
    s = heapsort_get (k);

    if (space[idx[k][s]][dim_tm] ||
        space[idx[k][s]][dim_3ds]) continue;

    /* get the description and print the line. */
    fprintf (fout, WEIGHTS_ENTRY, j % 2 ? "trGrey" : "trWhite",
             uids[idx[k][s]], uids[idx[k][s]],
             Wd[s] > ddev ? " color: rgb(0,100,0);" : "", Wd[s],
             Ws[s] > sdev ? " color: rgb(0,100,0);" : "", Ws[s]);

    j++;
  }

  /* close the output file. */
  fprintf (fout, WEIGHTS_FOOTER);
  fclose (fout);

  /* free the weights array. */
  free (Ws);
  free (Wd);
  free (sorted);

  /* free the clustering indices. */
  free (indices);
  for (k = 0; k < K; k++) free (idx[k]);
  free (n_idx);
  free (idx);

  /* exit the program. */
  return 0;
#else
  /* fail out. */
  fprintf (stderr, "FAIL! %s can't run without a mysql connection!\n", argv[0]);
  return 1;
#endif
}
