
/*
 * borg-classifier/stat.c: miscellaneous statistical functions source code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* stat_mean: compute means of selected points. */
double *stat_mean (unsigned long *idx, unsigned long n) {
  /* declare a value to hold the mean of each dimension. */
  double *mean;

  /* allocate memory for the mean. */
  mean = (double *) malloc (sizeof (double) * D);

  /* loop through the dimensions. */
  unsigned long i, j;
  for (j = 0; j < D; j++) {
    /* init the mean in this dimension. */
    mean[j] = 0.0;

    /* loop through the points. */
    for (i = 0; i < n; i++) {
      /* sum in the value in space here. */
      mean[j] += (double) space[idx[i]][j];
    }

    /* divide by the number of points to find the mean. */
    mean[j] /= (double) n;
  }

  /* return the mean. */
  return mean;
}

/* stat_var: compute bias-corrected sample variances of selected points. */
double *stat_var (double *mean, unsigned long *idx, unsigned long n) {
  /* declare a value to hold the variance of each dimension. */
  double *var;

  /* allocate memory for the variance. */
  var = (double *) malloc (sizeof (double) * D);

  /* loop through the dimensions. */
  unsigned long i, j;
  for (j = 0; j < D; j++) {
    /* init the variance in this dimension. */
    var[j] = 0.0;

    /* loop through the points. */
    for (i = 0; i < n; i++) {
      /* sum in the square of the deviation. */
      var[j] += pow (((double) space[idx[i]][j]) - mean[j], 2.0);
    }

    /* divide by one less the number of points to find the variance. */
    var[j] /= (double) (n - 1);
  }

  /* return the variances. */
  return var;
}

/* stat_cov: compute covariance matrix of two sets of selected points. */
double **stat_cov (double *amean, double *bmean,
                   unsigned long *aidx, unsigned long *bidx,
                   unsigned long n) {
  /* declare a matrix to hold the covariances. */
  unsigned long i, j, k;
  double **cov;

  /* allocate memory for the covariances. */
  cov = (double **) malloc (sizeof (double *) * D);

  /* loop through the rows of the covariance matrix. */
  for (i = 0; i < D; i++) {
    /* allocate this row of the matrix. */
    cov[i] = (double *) malloc (sizeof (double) * D);

    /* loop through the columns of the covariance matrix. */
    for (j = 0; j < D; j++) {
      /* initialize the covariance. */
      cov[i][j] = 0.0;

      /* loop through the points. */
      for (k = 0; k < n; k++) {
        /* sum the products of the deviations up. */
        cov[i][j] += ((double) space[aidx[k]][i] - amean[i]) *
                     ((double) space[bidx[k]][j] - bmean[j]);
      }

      /* divide to make this a covariance. */
      cov[i][j] /= (double) (n - 1);
    }
  }

  /* return the covariance matrix. */
  return cov;
}

/* stat_cholesky_decomp: returns the matrix L such that A=L*L'. */
double **stat_cholesky_decomp (double **A, unsigned long n) {
  /* declare required variables. */
  signed long i, j, k;
  double **L, sum;

  /* allocate memory for the L matrix. */
  L = (double **) malloc (sizeof (double *) * n);
  for (i = 0; i < n; i++) {
    /* allocate memory for this row. */
    L[i] = (double *) malloc (sizeof (double) * n);

    /* copy the values from A into this row. */
    for (j = 0; j < n; j++)
      L[i][j] = A[i][j];
  }

  /* perform the decomposition. */
  for (i = 0; i < n; i++) {
    for (j = i; j < n; j++) {
      /* generate a sum value. */
      for (sum = L[i][j], k = i - 1; k >= 0; k--)
        sum -= L[i][k] * L[j][k];

      /* for the diagonals... */
      if (i == j) {
        /* check for a positive definite sum. */
        if (sum <= 0.0) {
          /* print an error and fail. */
          fprintf (stderr, "WARN! cholesky decomposition failed!\n");
        /*fprintf (stderr, " input matrix:\n");
          for (i = 0; i < n; i++) {
            for (j = 0; j < n; j++) fprintf (stderr, "%12.6lf", A[i][j]);
            fprintf (stderr, "\n");
          }*/

          /* return nothing. */
          return NULL;
        }

        /* save the summed value. */
        L[i][i] = sqrt (sum);
      }
      else {
        /* save the off-diagonal summed value. */
        L[j][i] = sum / L[i][i];
      }
    }
  }

  /* zero the bottom left half. */
  for (i = 0; i < n; i++)
    for (j = 0; j < i; j++)
      L[j][i] = 0.0;

  /* return the new matrix. */
  return L;
}

/* stat_cholesky_solve: returns the solution (y) to A*y=b. */
double *stat_cholesky_solve (double **L, unsigned long n, double *b) {
  /* declare required variables. */
  unsigned int i, j;
  double *y, sum;

  /* allocate memory for the result. */
  y = (double *) malloc (sizeof (double) * n);

  /* loop through the rows. */
  for (i = 0; i < n; i++) {
    /* loop through the columns. */
    for (sum = b[i], j = 0; j < i; j++)
      sum -= L[i][j] * y[j];

    /* save the summed value. */
    y[i] = sum / L[i][i];
  }

  /* return the result. */
  return y;
}

/* stat_cholesky_inv: returns the inverse of A, given A=L*L'. */
double **stat_cholesky_inv (double **L, unsigned long n) {
  /* declare the required variables. */
  signed long i, j, k;
  double **C, sum;

  /* allocate memory for the output matrix. */
  C = (double **) malloc (sizeof (double *) * n);
  for (i = 0; i < n; i++) {
    /* allocate memory for the row. */
    C[i] = (double *) malloc (sizeof (double) * n);
    for (j = 0; j < n; j++)
      C[i][j] = 0.0;
  }

  /* run through the first block. */
  for (i = 0; i < n; i++) {
    /* in the rows. */
    for (j = 0; j <= i; j++) {
      /* calculate something. :P */
      sum = (i == j ? 1.0 : 0.0);
      for (k = i - 1; k >= j; k--) sum -= L[i][k] * C[j][k];
      C[j][i] = sum / L[i][i];
    }
  }

  /* and the second block. */
  for (i = n - 1; i >= 0; i--) {
    /* in the rows. */
    for (j = 0; j <= i; j++) {
      /* and something else. :P */
      sum = (i < j ? 0.0 : C[j][i]);
      for (k = i + 1; k < n; k++) sum -= L[k][i] * C[j][k];
      C[i][j] = C[j][i] = sum / L[i][i];
    }
  }

  /* return the inverted matrix. */
  return C;
}

/* stat_cholesky_logdet: returns the logarithm of |A|, given A=L*L'. */
double stat_cholesky_logdet (double **L, unsigned long n) {
  /* declare the log-determinant sum. */
  unsigned int i;
  double logdet = 0.0;

  /* loop through the diagonal. */
  for (i = 0; i < n; i++)
    logdet += log (L[i][i]);

  /* return the log-determinant. */
  logdet *= 2.0;
  return logdet;
}

/* stat_cholesky_free: frees a matrix L allocated by stat_cholesky_decomp(). */
void stat_cholesky_free (double **L, unsigned long n) {
  /* loop through the rows. */
  unsigned int i;
  for (i = 0; i < n; i++)
    free (L[i]);

  /* free the matrix. */
  free (L);
}

/* stat_qr: economy-sized qr-decomposition of a matrix. */
double **stat_qr_econ (double **A, unsigned long m, unsigned long n) {
  /* declare required variables. */
  unsigned long i, j;
  gsl_matrix *QR, *Q, *R;
  gsl_vector *tau;
  double **B;

  /* allocate memory for the output matrix. */
  B = (double **) malloc (sizeof (double *) * m);
  for (i = 0; i < m; i++)
    B[i] = (double *) malloc (sizeof (double) * (n < m ? n : m));

  /* allocate memory for the gsl elements. */
  QR = gsl_matrix_calloc (m, n);
  Q = gsl_matrix_calloc (m, m);
  R = gsl_matrix_calloc (m, n);
  tau = gsl_vector_calloc (m < n ? m : n);

  /* set the values of the QR input matrix. */
  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      gsl_matrix_set (QR, i, j, A[i][j]);

  /* decompose the matrix and unpack the orthogonal matrix. */
  gsl_linalg_QR_decomp (QR, tau);
  gsl_linalg_QR_unpack (QR, tau, Q, R);

  /* store the results. */
  for (i = 0; i < m; i++)
    for (j = 0; j < (n < m ? n : m); j++)
      B[i][j] = gsl_matrix_get (Q, i, j);

  /* free the gsl elements. */
  gsl_matrix_free (QR);
  gsl_matrix_free (Q);
  gsl_matrix_free (R);
  gsl_vector_free (tau);

  /* return the result. */
  return B;
}

/* stat_eigen: returns the matrix of eigenvalues of a given matrix. */
double **stat_eigen (double **A, unsigned long m, unsigned long n) {
  /* declare required variables. */
  unsigned long i, j;
  double **E;

  /* set up the output matrix. */
  E = (double **) malloc (sizeof (double *) * m);
  for (i = 0; i < m; i++)
    E[i] = (double *) malloc (sizeof (double) * n);

  /* set up the eigenvalues and eigenvectors. */
  gsl_vector *eval = gsl_vector_calloc (m);
  gsl_matrix *evec = gsl_matrix_calloc (m, n);

  /* set up the input matrix. */
  gsl_matrix *M = gsl_matrix_calloc (m, n);
  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      gsl_matrix_set (M, i, j, A[i][j]);

  /* set up the eigenworkspace. :P */
  gsl_eigen_symmv_workspace *work = gsl_eigen_symmv_alloc (m);

  /* calculate the eigenvectors. */
  gsl_eigen_symmv (M, eval, evec, work);
  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_DESC); {
    for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
        E[i][j] = gsl_matrix_get (evec, i, j);
  }

  /* free stuff. */
  gsl_vector_free (eval);
  gsl_matrix_free (evec);
  gsl_matrix_free (M);
  gsl_eigen_symmv_free (work);

  /* return the output matrix. */
  return E;
}

/* stat_matfree: frees memory associated with a matrix. */
void stat_matfree (double **A, unsigned long m) {
  /* declare required variables. */
  unsigned long i;

  /* free the rows. */
  for (i = 0; i < m; i++) free (A[i]);

  /* free the matrix. */
  free (A);
}

/* stat_matmult: returns the multiplication of two matrices. */
double **stat_matmult (double **A, unsigned long ma, unsigned long na,
                       double **B, unsigned long mb, unsigned long nb) {
  /* declare required variables. */
  unsigned long i, j, k;
  double **C;

  /* make sure we have the right dimensions. */
  if (na != mb) {
    /* fail. */
    fprintf (stderr, "FAIL! matrices do not have same inner dimensions!\n");
    exit (1);
  }

  /* allocate memory for the new matrix. */
  C = (double **) malloc (sizeof (double *) * ma);
  for (i = 0; i < ma; i++)
    C[i] = (double *) malloc (sizeof (double) * nb);

  /* do the multiplication. */
  for (i = 0; i < ma; i++) {
    for (j = 0; j < nb; j++) {
      for (C[i][j] = 0.0, k = 0; k < na; k++) {
        C[i][j] += A[i][k] * B[k][j];
      }
    }
  }

  /* return the new matrix. */
  return C;
}

/* stat_transpose: returns the transpose of the matrix A. */
double **stat_transpose (double **A, unsigned long m, unsigned long n) {
  /* declare required variables. */
  unsigned long i, j;
  double **B;

  /* allocate memory for the new matrix. */
  B = (double **) malloc (sizeof (double) * n);
  for (j = 0; j < n; j++)
    B[j] = (double *) malloc (sizeof (double) * m);

  /* loop through the new matrix. */
  for (i = 0; i < m; i++) {
    /* and through the rows, setting the value. */
    for (j = 0; j < n; j++)
      B[j][i] = A[i][j];
  }

  /* return the new matrix. */
  return B;
}

/* stat_inverse: returns the inverse of the matrix A. */
double **stat_inverse (double **A, unsigned long m, unsigned long n) {
  /* declare required variables. */
  unsigned int i, j;
  double **Ainv;
  int sig;

  /* allocate memory elements. */
  Ainv = (double **) malloc (sizeof (double *) * m);
  for (i = 0; i < m; i++)
    Ainv[i] = (double *) malloc (sizeof (double) * n);

  /* allocate gsl memory elements. */
  gsl_permutation *p = gsl_permutation_calloc (m);
  gsl_matrix *Ag = gsl_matrix_calloc (m, n);
  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      gsl_matrix_set (Ag, i, j, A[i][j]);

  /* calculate the LU decomposition. */
  gsl_linalg_LU_decomp (Ag, p, &sig);
  gsl_matrix *Ig = gsl_matrix_calloc (m, n);
  gsl_linalg_LU_invert (Ag, p, Ig);

  /* save the values. */
  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      Ainv[i][j] = gsl_matrix_get (Ig, i, j);

  /* return the inverted matrix. */
  gsl_permutation_free (p);
  gsl_matrix_free (Ag);
  gsl_matrix_free (Ig);
  return Ainv;
}

/* stat_cluster_disp: returns the dispersion of a given clustering. */
double stat_clustering_disp (double **X, unsigned long m, unsigned long n,
                             unsigned long *indices) {
  /* declare required variables. */
  unsigned long **idx, *n_idx;
  unsigned long ia, ib, j, k, K;
  double W, D, d;

  /* loop through the index array to find the number of clusters. */
  for (K = 0, ia = 0; ia < N; ia++) if (indices[ia] > K) K = indices[ia];
  K++;

  /* allocate some memory. */
  idx = (unsigned long **) malloc (sizeof (unsigned long *) * K);
  n_idx = (unsigned long *) malloc (sizeof (unsigned long) * K);
  for (k = 0; k < K; k++) {
    /* allocate the cluster indices. */
    idx[k] = (unsigned long *) malloc (sizeof (unsigned long) * N);
    n_idx[k] = 0;
  }

  /* loop through the points. */
  for (ia = 0; ia < N; ia++) {
    /* store the cluster in its index. */
    k = indices[ia];
    idx[k][n_idx[k]] = ia;
    n_idx[k]++;
  }

  /* loop through the clusters. */
  for (W = 0.0, k = 0; k < K; k++) {
    /* loop through the points. */
    for (D = 0.0, ia = 0; ia < n_idx[k]; ia++) {
      /* and again through the points. */
      for (ib = 0; ib < n_idx[k]; ib++) {
        /* don't uselessly compare the same point to itself. */
        if (ia == ib) continue;

        /* calculate the squared euclidean distance between the points. */
        for (d = 0.0, j = 0; j < n; j++)
          d += pow (X[idx[k][ia]][j] - X[idx[k][ib]][j], 2.0);

        /* add this to the total distance. */
        D += d;
      }
    }

    /* add the distance to the dispersion. */
    W += D / (2.0 * (double) n_idx[k]);
  }

  /* free the allocated indices. */
  for (k = 0; k < K; k++) free (idx[k]);
  free (n_idx);
  free (idx);

  /* return the variance. */
  return W;
}
