
/*
 * borg-classifier/util.c: base utility/init functions source code
 * Copyright (C) 2011 Bradley Worley
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * 
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
 */

/* core include. */
#include "core.h"

/* the input filenames. */
char *f_biuids, *f_bidims, *f_bispace;
char *f_uids, *f_dims, *f_space, *f_dist;

/* the association between database constants. */
#define N_CONST_ASSOC 10
#define CA_NAME 0
#define CA_PRE  1
#define CA_F    2
#define CA_TBL  3
#define CA_SQL  4
const char *db_const_assoc[N_CONST_ASSOC][5] = {
  { NAME_GOTERM,   PRE_GOTERM,   F_GOTERM,   TBL_GOTERM,   SQL_GOTERM },
  { NAME_KDISEASE, PRE_KDISEASE, F_KDISEASE, TBL_KDISEASE, SQL_KDISEASE },
  { NAME_KPATHWAY, PRE_KPATHWAY, F_KPATHWAY, TBL_KPATHWAY, SQL_KPATHWAY },
  { NAME_KEYWORD,  PRE_KEYWORD,  F_KEYWORD,  TBL_KEYWORD,  SQL_KEYWORD },
  { NAME_SEQSIM,   PRE_SEQSIM,   F_SEQSIM,   TBL_SEQSIM,   SQL_SEQSIM },
  { NAME_PFAM,     PRE_PFAM,     F_PFAM,     TBL_PFAM,     SQL_PFAM },
  { NAME_NOG,      PRE_NOG,      F_NOG,      TBL_NOG,      SQL_NOG },
  { NAME_PDBHET,   PRE_PDBHET,   F_PDBHET,   TBL_PDB_LIG,  SQL_PDBHET },
  { NAME_OMA,      PRE_OMA,      F_OMA,      TBL_OMA,      SQL_OMA },
  { NAME_BINDING,  PRE_BINDING,  F_BINDING,  TBL_BINDING,  SQL_BINDING }
};

/* the query pattern string to retrieve dimension names. */
#define MYSQL_QDIM "select distinct %s as %s from %s;"

/* file_exists: returns whether a file exists.
 * @fname: the filename to check existence on.
 */
int file_exists (const char *fname) {
  /* declare needed variables. */
  FILE *fh;

  /* try to open the file. */
  fh = fopen (fname, "r");

  /* check if the file exists. */
  if (fh) {
    /* close the file and return. */
    fclose (fh);
    return 1;
  }

  /* nope. */
  return 0;
}

/* init_uids: query the database for uniprot id information. */
void init_uids (void) {
  /* make some declarations. */
  FILE *fh = NULL;
  char fline[256], *fp;
  unsigned long i;

  /* zero the uid count. */
  N = 0;

  /* see if we have already written the file. */
  if (file_exists (f_uids)) {
    /* read in the uids file. */
    fh = fopen (f_uids, "rb");
    while (!feof (fh)) {
      /* get a line from the file. */
      if (fgets (fline, 256, fh) == NULL) continue;
      fp = strchr (fline, ',');
      fp++; i = 0;
      while (i < strlen (fp)) {
        if (fp[i] == '\n') fp[i] = '\0';
        i++;
      }

      /* add the uid to the array. */
      uids = (char **) realloc (uids, sizeof (char *) * (N + 1));
      uids[N] = str_dup (fp);
      N++;
    }

    /* quit the function early. */
    fprintf (stderr, "INFO: read in %lu uniprot identifiers.\n", N);
    return;
  }

#if CONFIGURE_MYSQL == 1
  /* try the uniprot query. */
  if (db_query ("select distinct UID from " TBL_UNIPROT ";")) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query uniprot ids!\n");
    exit (1);
  }

  /* open the uids file. */
  fh = fopen (f_uids, "wb");

  /* get the query results. */
  struct db_results *results = db_results_get ();

  /* allocate memory for the uniprot array. */
  N = results->rows;
  fprintf (stderr, "INFO: queried %lu uniprot identifiers.\n", N);
  uids = (char **) malloc (sizeof (char *) * N);

  /* loop through the entries in the uniprot query. */
  for (i = 0; i < N; i++) {
    /* save the field value. */
    uids[i] = str_dup (db_results_field (results, i, NAME_UID));
    fprintf (fh, "%lu,%s\n", i, uids[i]);
  }

  /* free the query results. */
  db_results_free (results);

  /* close the uids file. */
  fclose (fh);
#else
  /* print an error message and exit. */
  fprintf (stderr, "FAIL! no uids file and no mysql connection!\n");
  exit (1);
#endif
}

/* init_biuids: initializes the list of uids found in both datasets. */
void init_biuids (void) {
  /* make some declarations. */
  FILE *fh = NULL;
  char fline[256], fdb[256], fuid[256];
  unsigned long i, j, k, found;
  struct db_results *results;

  /* get the biuids filename. */
  conf_check (CONF_BI_UIDS);
  f_biuids = conf_get (CONF_BI_UIDS);

  /* zero out the biuid count. */
  M = 0;
  biuids_db = NULL;

  /* see if we have already written to the file. */
  if (file_exists (f_biuids)) {
    /* read in the biuids file. */
    fh = fopen (f_biuids, "rb");
    while (!feof (fh)) {
      /* get the next line. */
      if (fgets (fline, 256, fh)) {
        if (sscanf (fline, "%lu,%[^,],%[^ ] ", &i, fdb, fuid) == 3) {
          /* add the uid to the array. */
          biuids = (char **) realloc (biuids, sizeof (char *) * (M + 1));
          for (k = strlen (fuid) - 1; k > 1; k--)
            if (fuid[k] == '\n') fuid[k] = '\0';
          biuids[M] = str_dup (fuid);

          /* add the database to the array. */
          biuids_db = (char **) realloc (biuids_db, sizeof (char *) * (M + 1));
          biuids_db[M] = str_dup (fdb);

          /* increment the counter. */
          M++;
        }
      }
    }

    /* quit the function early. */
    fprintf (stderr, "INFO: read in %lu joined uniprot identifiers.\n", M);
    return;
  }

#if CONFIGURE_MYSQL == 1
  /* connect to the borg_np database. */
  db_connect (DB_BORG_NP);

  /* try the uniprot query. */
  if (db_query ("select distinct UID from " TBL_UNIPROT ";")) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query uniprot ids!\n");
    exit (1);
  }

  /* get the query results. */
  results = db_results_get ();

  /* allocate memory for the uniprot array. */
  M = results->rows;
  fprintf (stderr, "INFO: queried %lu NP uniprot identifiers.\n", M);
  biuids = (char **) malloc (sizeof (char *) * M);
  biuids_db = (char **) malloc (sizeof (char *) * M);

  /* loop through the entries in the uniprot query. */
  for (i = 0; i < M; i++) {
    /* save the field value. */
    biuids[i] = str_dup (db_results_field (results, i, NAME_UID));
    biuids_db[i] = str_dup (DB_BORG_NP);
  }

  /* free the query results. */
  db_results_free (results);

  /* disconnect. */
  db_disconnect ();

  /* connect to the borg_pdac database. */
  db_connect (DB_BORG_PDAC);

  /* try the uniprot query. */
  if (db_query ("select distinct UID from " TBL_UNIPROT ";")) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query uniprot ids!\n");
    exit (1);
  }

  /* get the query results. */
  results = db_results_get ();

  /* allocate memory for the uniprot array. */
  fprintf (stderr, "INFO: queried %lu PDAC uniprot identifiers.\n",
    results->rows);

  /* loop through the entries in the uniprot query. */
  for (i = 0; i < results->rows; i++) {
    /* see if it was already found in NP. */
    found = 0;
    for (j = 0; j < M; j++)
      if (strcmp (biuids[j], db_results_field (results, i, NAME_UID)) == 0)
        found = 1;

    /* save the field value. */
    if (!found) {
      /* resize the array and increment the counter. */
      M++;
      biuids = (char **) realloc (biuids, sizeof (char *) * M);
      biuids_db = (char **) realloc (biuids_db, sizeof (char *) * M);

      /* save the value. */
      biuids[M - 1] = str_dup (db_results_field (results, i, NAME_UID));
      biuids_db[M - 1] = str_dup (DB_BORG_PDAC);
    }
  }

  /* free the query results. */
  db_results_free (results);

  /* disconnect. */
  db_disconnect ();

  /* open the uids file. */
  fh = fopen (f_biuids, "wb");

  /* loop through the biuids. */
  for (i = 0; i < M; i++)
    fprintf (fh, "%lu,%s,%s\n", i, biuids_db[i], biuids[i]);

  /* close the uids file. */
  fclose (fh);
#else
  /* print an error message and exit. */
  fprintf (stderr, "FAIL! no uids file and no mysql connection!\n");
  exit (1);
#endif
}

/* init_dim: query a single annotation type. */
char **init_dim (const char *name, const char *query, unsigned long *count) {
  /* declare required variables. */
  struct db_results *results;
  unsigned long i, n;
  char **d;

  /* init the count. */
  *count = 0;

  /* try the goterm query. */
  if (db_query (query)) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query %s dimensions!\n", name);
    exit (1);
  }

  /* get the query results. */
  results = db_results_get ();

  /* allocate memory for the query. */
  n = results->rows;
  fprintf (stderr, "INFO: queried %lu %s dims.\n", n, name);
  d = (char **) malloc (sizeof (char *) * n);

  /* loop through the entries in the query. */
  for (i = 0; i < n; i++) {
    /* save the field value. */
    d[i] = str_dup (db_results_field (results, i, name));
  }

  /* free the query results. */
  db_results_free (results);

  /* set the count pointer. */
  *count = n;

  /* return the array. */
  return d;
}

/* init_dimset: initialize a single set of dimensions for dims. */
void init_dimset (FILE *fh, unsigned long k) {
  /* declare required variables. */
  unsigned long i, ntmp;
  char **dtmp, query[1024];

  /* should we pull in term terms? */
  if (conf_null (CONF_DIMSET) ||
      strstr (conf_get (CONF_DIMSET),db_const_assoc[k][CA_NAME])) {
    /* build the query string. */
    snprintf (query, 256, MYSQL_QDIM,
              db_const_assoc[k][CA_F],
              db_const_assoc[k][CA_NAME],
              db_const_assoc[k][CA_TBL]);

    /* query the terms. */
    dtmp = init_dim (db_const_assoc[k][CA_NAME], query, &ntmp);

    /* allocate memory for the query. */
    if (dims) dims = (char **) realloc (dims, sizeof (char *) * (D + ntmp));
    else dims = (char **) malloc (sizeof (char *) * (D + ntmp));

    /* loop through the entries in the query. */
    for (i = 0; i < ntmp; i++) {
      /* save the field value. */
      dims[D] = str_dup (dtmp[i]);
      fprintf (fh, "%lu,%s\n", D, dims[D]);
      D++;
    }

    /* free the query results. */
    str_freev (dtmp, ntmp);
  }
}

/* init_bidimset: initialize a single set of dimensions for bidims. */
void init_bidimset (FILE *fh, unsigned long k) {
  /* declare required variables. */
  unsigned long i, j, n_np, n_pdac, found;
  char **d_np, **d_pdac, query[1024];

  /* should we pull in term terms? */
  if (conf_null (CONF_DIMSET) ||
      strstr (conf_get (CONF_DIMSET), db_const_assoc[k][CA_NAME])) {
    /* build the query string. */
    snprintf (query, 256, MYSQL_QDIM,
              db_const_assoc[k][CA_F],
              db_const_assoc[k][CA_NAME],
              db_const_assoc[k][CA_TBL]);

    /* query the terms. */
    db_connect (DB_BORG_NP);
    d_np = init_dim (db_const_assoc[k][CA_NAME], query, &n_np);
    db_disconnect ();
    db_connect (DB_BORG_PDAC);
    d_pdac = init_dim (db_const_assoc[k][CA_NAME], query, &n_pdac);

    /* loop through each NP dimension. */
    for (i = 0; i < n_np; i++) {
      /* try to find the NP dimension in the PDAC dimset. */
      for (found = 0, j = 0; j < n_pdac; j++)
        if (strcmp (d_np[i], d_pdac[j]) == 0)
          found = 1;

      /* only add if we found it in both sets. */
      if (found) {
        /* increment the counter. */
        R++;

        /* allocate memory for the query. */
        if (bidims) bidims = (char **) realloc (bidims, sizeof (char *) * R);
        else bidims = (char **) malloc (sizeof (char *) * R);

        /* add the new bidim, save to the file. */
        bidims[R - 1] = str_dup (d_np[i]);
        fprintf (fh, "%lu,%s\n", R - 1, bidims[R - 1]);
      }
    }

    /* free the query results. */
    str_freev (d_np, n_np);
    str_freev (d_pdac, n_pdac);
  }
}

/* init_dims: query the database for dimensions information. */
void init_dims (void) {
  /* make some declarations. */
  FILE *fh = NULL;
  char fline[256], *fp;
  unsigned long i;

  /* zero the dimension count. */
  D = 0;

  /* see if we have already written the file. */
  if (file_exists (f_dims)) {
    /* read in the dims file. */
    fh = fopen (f_dims, "rb");
    while (!feof (fh)) {
      /* get a line from the file. */
      if (fgets (fline, 256, fh) == NULL) continue;
      fp = strchr (fline, ',');
      fp++; i = 0;
      while (i < strlen (fp)) {
        if (fp[i] == '\n') fp[i] = '\0';
        i++;
      }

      /* add the uid to the array. */
      dims = (char **) realloc (dims, sizeof (char *) * (D + 1));
      dims[D] = str_dup (fp);
      D++;
    }

    /* quit the function early. */
    fprintf (stderr, "INFO: read in %lu total dimensions.\n", D);
    return;
  }

#if CONFIGURE_MYSQL == 1
  /* open the dimensions file. */
  fh = fopen (f_dims, "wb");

  /* get the set of dimensions to pull in. */
  conf_check (CONF_DIMSET);

  if (conf_null (CONF_DIMSET) || strstr (conf_get (CONF_DIMSET), NAME_REG)) {
    /* add the regulation dimensions. */
    D += 2;
    dims = (char **) malloc (sizeof (char *) * D);
    dims[0] = str_dup (F_REG_UP);
    fprintf (fh, "%lu,%s\n", 0L, dims[0]);
    dims[1] = str_dup (F_REG_DOWN);
    fprintf (fh, "%lu,%s\n", 1L, dims[1]);
  }

  /* loop through the available dimensions. */
  for (i = 0; i < N_CONST_ASSOC; i++)
    init_dimset (fh, i);

  /* close the dimensions file. */
  fclose (fh);

  /* print the total dimensions. */
  fprintf (stderr, "INFO: queried %lu total dimensions.\n", D);
#else
  /* print an error message and exit. */
  fprintf (stderr, "FAIL! no dims file and no mysql connection!\n");
  exit (1);
#endif
}

/* init_bidims: initializes the list of dims found in both datasets. */
void init_bidims (void) {
  /* make some declarations. */
  FILE *fh = NULL;
  char fline[256], *fp;
  unsigned long i;

  /* get the bidims filename. */
  conf_check (CONF_BI_DIMS);
  f_bidims = conf_get (CONF_BI_DIMS);

  /* zero the dimension count. */
  R = 0;

  /* see if we have already written the file. */
  if (file_exists (f_bidims)) {
    /* read in the dims file. */
    fh = fopen (f_bidims, "rb");
    while (!feof (fh)) {
      /* get a line from the file. */
      if (fgets (fline, 256, fh) == NULL) continue;
      fp = strchr (fline, ',');
      fp++; i = 0;
      while (i < strlen (fp)) {
        if (fp[i] == '\n') fp[i] = '\0';
        i++;
      }

      /* add the uid to the array. */
      bidims = (char **) realloc (bidims, sizeof (char *) * (R + 1));
      bidims[R] = str_dup (fp);
      R++;
    }

    /* quit the function early. */
    fprintf (stderr, "INFO: read in %lu total joined dimensions.\n", R);
    return;
  }

#if CONFIGURE_MYSQL == 1
  /* open the dimensions file. */
  fh = fopen (f_bidims, "wb");

  /* get the set of dimensions to pull in. */
  conf_check (CONF_DIMSET);

  if (conf_null (CONF_DIMSET) || strstr (conf_get (CONF_DIMSET), NAME_REG)) {
    /* add the regulation dimensions. */
    R += 2;
    bidims = (char **) malloc (sizeof (char *) * R);
    bidims[R - 2] = str_dup (F_REG_UP);
    bidims[R - 1] = str_dup (F_REG_DOWN);
    fprintf (fh, "%lu,%s\n", R - 2, bidims[R - 2]);
    fprintf (fh, "%lu,%s\n", R - 1, bidims[R - 1]);
  }

  if (conf_null (CONF_DIMSET) || strstr (conf_get (CONF_DIMSET), NAME_CANCER)) {
    /* add the cancer dimensions. */
    R++;
    if (bidims) bidims = (char **) realloc (bidims, sizeof (char *) * R);
    else bidims = (char **) malloc (sizeof (char *) * R);
    bidims[R - 1] = str_dup (NAME_CANCER);
    fprintf (fh, "%lu,%s\n", R - 1, bidims[R - 1]);
  }

  /* loop through the available dimensions. */
  for (i = 0; i < N_CONST_ASSOC; i++)
    init_bidimset (fh, i);

  /* close the dimensions file. */
  fclose (fh);

  /* print the total dimensions. */
  fprintf (stderr, "INFO: queried %lu total joined dimensions.\n", R);
#else
  /* print an error message and exit. */
  fprintf (stderr, "FAIL! no bidims file and no mysql connection!\n");
  exit (1);
#endif
}

/* find_knowns: populates the arrays of known protein indices. */
#if CONFIGURE_MYSQL == 1
void find_knowns (void) {
  /* allocate memory. */
  unsigned long i;
  fprintf (stderr, "INFO: allocating memory for known proteins.\n");
  if (db_query (Q_MMP)) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query for knowns!\n"); exit (1);
  }

  /* get the query results. */
  struct db_results *results = db_results_get ();

  /* allocate memory for the uniprot array. */
  n_mmp = results->rows;
  fprintf (stderr, "INFO: queried %lu matrix metalloproteins.\n", n_mmp);
  mmp = (unsigned long *) malloc (sizeof (unsigned long) * n_mmp);

  /* loop through the entries in the uniprot query. */
  for (i = 0; i < n_mmp; i++)
    mmp[i] = uid_index (db_results_field (results, i, NAME_UID));

  /* free the query results. */
  db_results_free (results);

  /* topoisomerases. */
  if (db_query (Q_TOP)) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query for knowns!\n"); exit (1);
  }

  /* get the query results. */
  results = db_results_get ();

  /* allocate memory for the uniprot array. */
  n_top = results->rows;
  fprintf (stderr, "INFO: queried %lu topoisomerases.\n", n_top);
  top = (unsigned long *) malloc (sizeof (unsigned long) * n_top);

  /* loop through the entries in the uniprot query. */
  for (i = 0; i < n_top; i++)
    top[i] = uid_index (db_results_field (results, i, NAME_UID));

  /* free the query results. */
  db_results_free (results);

  /* egf. */
  if (db_query (Q_EGF)) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query for knowns!\n"); exit (1);
  }

  /* get the query results. */
  results = db_results_get ();

  /* allocate memory for the uniprot array. */
  n_egf = results->rows;
  fprintf (stderr, "INFO: queried %lu epidermal growth factors.\n", n_egf);
  egf = (unsigned long *) malloc (sizeof (unsigned long) * n_egf);

  /* loop through the entries in the uniprot query. */
  for (i = 0; i < n_egf; i++)
    egf[i] = uid_index (db_results_field (results, i, NAME_UID));

  /* free the query results. */
  db_results_free (results);
}

/* init_uid_termset: initializes a set of terms in a single uid vector. */
void init_uid_termset (unsigned long uid,
                       const char *query,
                       const char *field) {
  /* make sure the dimension was selected. */
  if (strcmp (field, F_REG_UP) &&
      strcmp (field, F_REG_DOWN) &&
      !conf_null (CONF_DIMSET) &&
      !strstr (conf_get (CONF_DIMSET), field))
    return;

  /* try the goterm protein query. */
  if (db_query (query, uids[uid], "")) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query go terms for %s!\n", uids[uid]);
    exit (1);
  }

  /* get the query results. */
  struct db_results *results = db_results_get ();

  /* declare loop and matrix indices. */
  signed int fv;
  unsigned long i, j, dim;
  char *lbl, *sfv;

  /* loop through the entries in the query. */
  for (i = 0; i < results->rows; i++) {
    /* are we doing regulations? */
    if (strcmp (field, NAME_REG_UP) == 0) {
      /* set the value. */
      sfv = db_results_field (results, i, field);
      if (sfv) fv = atol (sfv);
      if (sfv && fv) space[uid][0] = 1;
    }
    else if (strcmp (field, NAME_REG_DOWN) == 0) {
      /* set the value. */
      sfv = db_results_field (results, i, field);
      if (sfv) fv = atol (sfv);
      if (sfv && fv) space[uid][1] = 1;
    }
    else {
      /* set the default dimension index. */
      dim = D + 16;

      /* get the dimension label. */
      lbl = db_results_field (results, i, field);

      /* get the dimension index. */
      for (j = 0; j < D; j++) {
        /* see if we have a match. */
        if (strcmp (dims[j], lbl) == 0) {
          /* set the index and break. */
          dim = j;
          break;
        }
      }

      /* see if we found it. */
      if (dim >= D) {
        /* fail out. */
        fprintf (stderr, "FAIL! dimension %s not found!\n", lbl);
        exit (1);
      }

      /* set the new value. */
      space[uid][dim] = 1;
    }
  }

  /* free the query results. */
  db_results_free (results);
}

/* init_uid: initializses the vector along a single uid. */
void init_uid (unsigned long uid) {
  /* allocate the uid vector. */
  space[uid] = (signed int *) malloc (sizeof (signed int) * D);

  /* get the starting clock count. */
  static time_t start, finish;
  if (uid == 0) {
    /* get the starting clock count. */
    time (&start);
  }

  /* zero out the row. */
  unsigned long i;
  for (i = 0; i < D; i++)
    space[uid][i] = 0;

  /* initialize the different term sets. */
  init_uid_termset (uid, SQL_GOTERM, NAME_GOTERM);
  init_uid_termset (uid, SQL_KDISEASE, NAME_KDISEASE);
  init_uid_termset (uid, SQL_KPATHWAY, NAME_KPATHWAY);
  init_uid_termset (uid, SQL_KEYWORD, NAME_KEYWORD);
  init_uid_termset (uid, SQL_PFAM, NAME_PFAM);
  init_uid_termset (uid, SQL_NOG, NAME_NOG);
  init_uid_termset (uid, SQL_REG_UP, NAME_REG_UP);
  init_uid_termset (uid, SQL_REG_DOWN, NAME_REG_DOWN);

  /* see if we should output a message. */
  if (uid == 100) {
    /* get the ending clock count. */
    time (&finish);
    double est = ((double) finish - (double) start) /
                  100.0 * ((double) N - 100.0) / 60.0;
    fprintf (stderr, "INFO: you can probably grab a cup of coffee now,\n");
    fprintf (stderr, "INFO: (i guess it'll be another %.1lf minutes)\n", est);
  }
}

/* write_space: outputs the space contained by the uids and dims. */
void write_space (void) {
  /* open the space file. */
  FILE *fh = fopen (f_space, "wb");

  /* loop through the space. */
  unsigned long i, j;
  for (i = 0; i < N; i++) {
    /* loop through the dimensions. */
    fprintf (fh, "%lu,", i);
    for (j = 0; j < D; j++) {
      /* set the bit in the dim variable. */
      fprintf (fh, "%d", space[i][j]);
    }

    /* print the newline. */
    fprintf (fh, "\n");
  }

  /* close the space file. */
  fclose (fh);
}
#endif

/* read_space: inputs the space contained by the uids and dims. */
void read_space (void) {
  /* make some declarations. */
  FILE *fh = NULL;
  char fline[D * 2 + 2], dim[D * 2 + 2];
  unsigned long i, uid;

  /* see if we shouldn't read the space file. */
  if (conf_null (CONF_SPACE)) return;

  /* allocate the space matrix. */
  space = (signed int **) malloc (sizeof (signed int *) * N);

  /* read in the space file. */
  fh = fopen (f_space, "rb");
  while (!feof (fh)) {
    /* get a line from the file. */
    if (fgets (fline, D * 2 + 2, fh) == NULL) continue;
    if (sscanf (fline, "%lu,%s ", &uid, dim) == 2) {
      /* allocate the uid vector. */
      space[uid] = (signed int *) malloc (sizeof (signed int) * D);

      /* read in the dimension values. */
      for (i = 0; i < D; i++) {
        /* see how to flip the bit. */
        if (dim[i] == '1') space[uid][i] = 1;
        else space[uid][i] = 0;
      }
    }
  }

  /* output a debug message. */
  fprintf (stderr, "INFO: read in %lu spatial elements.\n", N * D);
}

/* init_space: initializes the space contained by the uids and dims. */
void init_space (void) {
  /* see if we should give up on the space file. */
  if (conf_null (CONF_SPACE)) return;

  /* see if we have the space file. */
  if (file_exists (f_space)) {
    /* read the file and return. */
    read_space ();
    return;
  }

#if CONFIGURE_MYSQL == 1
  /* print a debug message. */
  fprintf (stderr, "INFO: generating a space of %lu elements.\n", N * D);

  /* allocate the space matrix. */
  space = (signed int **) malloc (sizeof (signed int *) * N);
  if (!space) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to allocate memory for space matrix!\n");
    exit (1);
  }

  /* loop through the uids. */
  unsigned long i;
  for (i = 0; i < N; i++) {
    /* initialize the row at this uid. */
    init_uid (i);
  }

  /* write the space file. */
  fprintf (stderr, "INFO: writing generated spatial elements.\n");
  write_space ();
#else
  /* print an error message and exit. */
  fprintf (stderr, "FAIL! no space file and no mysql connection!\n");
  exit (1);
#endif
}

/* init_bispace_cell: initializes a single cell of the bispace matrix. */
void init_bispace_cell (unsigned long i, unsigned long j) {
  /* declare required variables. */
  struct db_results *results;
  unsigned long k;
  char pre[32], qtemplate[1024], qclause[1024];

  /* init vars. */
  k = 0;

  /* connect to the appropriate database. */
  db_connect (biuids_db[i]);

  /* check for boilerplate solutions. */
  if (strcmp (bidims[j], NAME_REG_UP) == 0) {
    /* use the reg:UP query. */
    strcpy (qtemplate, SQL_REG_UP);
    snprintf (qclause, 1024, "having reg_UP = '1'");
  }
  else if (strcmp (bidims[j], NAME_REG_DOWN) == 0) {
    /* use the reg:UP query. */
    strcpy (qtemplate, SQL_REG_DOWN);
    snprintf (qclause, 1024, "having reg_DOWN = '1'");
  }
  else if (strcmp (bidims[j], NAME_CANCER) == 0) {
    /* just base the answer on the database. */
    if (strcmp (biuids_db[i], DB_BORG_PDAC) == 0)
      bispace[i][j] = 1;
    else
      bispace[i][j] = 0;

    /* done. */
    return;
  }
  else {
    /* parse the prefix from the dimension. */
    if (sscanf (bidims[j], "%[^:]:", pre) != 1) {
      /* fail out. */
      fprintf (stderr, "FAIL! unable to parse prefix from dim %s of uid %s!\n",
               bidims[j], biuids[i]);

      /* exit. */
      return;
    }

    /* determine the annotation type (PRE_*) to query. */
    for (k = 0; k < N_CONST_ASSOC; k++) {
      /* check the matching. */
      if (strcmp (db_const_assoc[k][CA_PRE], pre) == 0) {
        /* we found it! */
        break;
      }
    }

    /* make sure we found an index. */
    if (k >= N_CONST_ASSOC) {
      /* fail out! */
      fprintf (stderr, "FAIL! unable to find prefix %s of dim %s!\n",
               pre, bidims[j]);

      /* exit. */
      return;
    }

    /* set up the query clause. */
    strcpy (qtemplate, db_const_assoc[k][CA_SQL]);
    snprintf (qclause, 1024, "having %s = '%s'", db_const_assoc[k][CA_NAME],
              bidims[j]);
  }

  /* query, with the 'having' clause, under the identified annotation type. */
  if (db_query (qtemplate, biuids[i], qclause)) {
    /* fail. */
    fprintf (stderr, "FAIL! unable to query dim %s for uid %s in %s!\n",
             bidims[j], biuids[i], biuids_db[i]);

    /* exit. */
    return;
  }

  /* get the query results. */
  results = db_results_get ();

  /* see if we got a result. (Xij=1, otherwise 0) */
  if (results->rows)
    bispace[i][j] = 1;
  else
    bispace[i][j] = 0;

  /* free the query results. */
  db_results_free (results);

  /* disconnect from the database. */
  db_disconnect ();
}

/* init_bispace: initializes the data matrix of biuids and bidims. */
void init_bispace (void) {
  /* declare required variables. */
  unsigned long i, j;
  FILE *fh;

  /* get the bispace filename. */
  conf_check (CONF_BI_SPACE);
  f_bispace = conf_get (CONF_BI_SPACE);

  /* see if the bispace file exists. */
  if (file_exists (f_bispace)) {
    /* print a kind message. */
    fprintf (stderr, "INFO: joined space file exists. cool beans.\n");
  }
  else {
    /* print a kind message. */
    fprintf (stderr, "INFO: building a joined space matrix of %lu elements.\n",
             M * R);

    /* allocate memory for the bispace matrix. */
    bispace = (signed int **) malloc (sizeof (signed int *) * M);
    for (i = 0; i < M; i++)
      bispace[i] = (signed int *) malloc (sizeof (signed int) * R);

    /* initialize the bispace matrix elements. */
    for (i = 0; i < M; i++)
      for (j = 0; j < R; j++)
        init_bispace_cell (i, j);

    /* print another kind message. */
    fprintf (stderr, "INFO: writing joined space matrix to disk.\n");
    fflush (stderr);

    /* write the bispace file as an SVMlight-parseable file. */
    fh = fopen (f_bispace, "wb");
    for (i = 0; i < M; i++) {
      /* just base the answer on the database. */
      if (strcmp (biuids_db[i], DB_BORG_PDAC) == 0)
        fprintf (fh, "+1");
      else
        fprintf (fh, "-1");

      /* and through the dims. */
      for (j = 0; j < R; j++)
        fprintf (fh, " %lu:%d", i + 1, bispace[i][j]);

      /* and a newline. */
      fprintf (fh, " # %s\n", biuids[i]);
    }
    fclose (fh);
  }
}

/* init_dist: calculates the distance matrix. */
void init_dist (void) {
  /* declare loop indices. */
  unsigned long i, j, k;
  signed int dist;
  static time_t start, finish;

  /* see if we shouldn't read the space file. */
  if (conf_null (CONF_DIST)) return;

  /* allocate the rows. */
  distances = (signed int **) malloc (sizeof (signed int *) * N);
  max_dist = 0;

  /* loop through the rows. */
  for (i = 0; i < N; i++) {
    /* allocate the columns. */
    distances[i] = (signed int *) malloc (sizeof (signed int) * N);
  }

  /* see if we have the distances file. */
  if (file_exists (f_dist)) {
    /* open the file. */
    FILE *fh = fopen (f_dist, "rb");

    /* loop through the file. */
    while (!feof (fh)) {
      /* try reading in a line. */
      if (fscanf (fh, "%lu,%lu,%d ", &i, &j, &dist) == 3)
        distances[i][j] = dist;
    }

    /* write a debug line. */
    fprintf (stderr, "INFO: read in %lu distance elements.\n", N * N);

    /* exit the function early. */
    fclose (fh);
    return;
  }

  /* record the start time. */
  fprintf (stderr, "INFO: calculating a distance matrix of %lu elements.\n",
           N * N);
  time (&start);

  /* loop through the rows. */
  for (i = 0; i < N; i++) {
    /* loop through the columns. */
    for (j = i; j < N; j++) {
      /* calculate the distance. */
      distances[i][j] = 0;
      for (k = 0; k < D; k++) {
        /* add the distance value. */
        if (space[i][k] != space[j][k]) distances[i][j]++;
      }

      /* check the max distance. */
      if (distances[i][j] > max_dist) max_dist = distances[i][j];

      /* fill in the other half. */
      if (j != i) distances[j][i] = distances[i][j];
    }

    /* see if we should print an eta. */
    if (i == 10) {
      /* get the ending clock count. */
      time (&finish);
      double est = ((double) finish - (double) start) /
                    10.0 * ((double) N - 10.0) / 60.0;
      fprintf (stderr, "INFO: didn't get the coffee? now's the time!\n");
      fprintf (stderr, "INFO: (i guess it'll be another %.1lf minutes)\n",
               est);
    }
  }

  /* print the maximal distance. */
  fprintf (stderr, "INFO: found a maximum distance of %d features.\n",
           max_dist);

  /* print a debug message. */
  fprintf (stderr, "INFO: writing calculated distance matrix.\n");

  /* loop through the matrix. */
  FILE *fh = fopen (f_dist, "wb");
  for (i = 0; i < N; i++) {
    /* and through the columns. */
    for (j = 0; j < N; j++) {
      /* write the distance line. */
      fprintf (fh, "%lu,%lu,%d\n", i, j, distances[i][j]);
    }
  }
  fclose (fh);
}

/* uid_index: gets the index of a specific uid in the array. */
unsigned long uid_index (const char *uid) {
  /* first, try looping through the list. */
  unsigned long idx = N + 16, i;
  for (i = 0; i < N; i++) {
    /* see if we have a match. */
    if (strcmp (uids[i], uid) == 0) {
      /* break out of the loop. */
      idx = i;
      break;
    }
  }

  /* see if we got it. */
  if (idx < N) return idx;

#if CONFIGURE_MYSQL == 1
  /* didn't find it. try an aliases query. */
  if (db_query ("select distinct orig from aliases where alias='%s';", uid)) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query aliases for %s!\n", uid);
    return idx;
  }

  /* get the query results. */
  unsigned long j;
  struct db_results *results = db_results_get ();

  /* loop through the results. */
  for (i = 0; i < results->rows; i++) {
    for (j = 0; j < N; j++) {
      /* see if we have a match. */
      if (strcmp (uids[j], db_results_field (results, i, "orig")) == 0) {
        /* break out of the loop. */
        idx = j;
        break;
      }
    }

    /* see if we found it. */
    if (idx < N) break;
  }

  /* free the query results. */
  db_results_free (results);

  /* print a warning message if required. */
  if (idx > N)
    fprintf (stderr,
             "WARN! tried to get the index of unindexed protein '%s'\n",
             uid);

  /* return the index value. */
  return idx;
#else
  /* print a warning message. */
  fprintf (stderr, "FAIL! no mysql connection to check aliases for %s!\n",
           uid);
  return idx;
#endif
}

/* dim_index: gets the index of a specific dimension in the array. */
unsigned long dim_index (const char *dim) {
  /* first, try looping through the list. */
  unsigned long idx = D + 16, j;
  for (j = 0; j < D; j++) {
    /* see if we have a match. */
    if (strcmp (dims[j], dim) == 0) {
      /* break out of the loop. */
      idx = j;
      break;
    }
  }

  /* see if we got it. */
  if (idx < D) return idx;

  /* print a warning message if required. */
  if (idx > D)
    fprintf (stderr,
             "WARN! tried to get the index of unindexed dimension '%s'\n",
             dim);

  /* return the index value. */
  return idx;
}

/* uid_has_pdb: see if a protein has a pdb associated with it. */
int uid_has_pdb (unsigned long i) {
#if CONFIGURE_MYSQL == 1
  /* make sure we're in bounds. */
  if (i >= N) {
    /* fail out. */
    fprintf (stderr, "FAIL! uid index '%lu' out of range!\n", i);
    exit (1);
  }

  /* query the database for pdb structures. */
  if (db_query ("select PDB from pdbProteins where UID = '%s';", uids[i])) {
    /* fail out. */
    fprintf (stderr, "FAIL! unable to query for pdb structures!\n");
    exit (1);
  }

  /* get the results. */
  struct db_results *results = db_results_get ();
  int has = 0; if (results->rows > 0) has = 1;

  /* free the results. */
  db_results_free (results);

  /* return the value. */
  return has;
#else
  /* print a failure message and exit. */
  fprintf (stderr, "FAIL! no mysql connection to check for pdb linkage!\n");
  exit (1);
#endif
}

/* dim_desc: gets the description text of a dimension. */
char *dim_desc (unsigned long j) {
#if CONFIGURE_MYSQL == 1
  /* declare a description string. */
  char *desc, *fval;
  int ret = 2;

  /* see what our dimension type is. */
  if (strstr (dims[j], "goterm:")) {
    /* set up the query. */
    ret = db_query ("select distinct dsc from " \
                    TBL_GOTERM " where " F_GOTERM " = '%s';", dims[j]);
  }
  else if (strstr (dims[j], "keggd:")) {
    /* set up the query. */
    ret = db_query ("select distinct dsc from " \
                    TBL_KDISEASE " where " F_KDISEASE " = '%s';", dims[j]);
  }
  else if (strstr (dims[j], "keggp:")) {
    /* set up the query. */
    ret = db_query ("select distinct dsc from " \
                    TBL_KPATHWAY " where " F_KPATHWAY " = '%s';", dims[j]);
  }
  else if (strstr (dims[j], "keyw:")) {
    /* set up the query. */
    return str_dup (dims[j]);
  }
  else if (strstr (dims[j], "pfamily:")) {
    /* set up the query. */
    ret = db_query ("select distinct dsc from " \
                    TBL_PFAM " where " F_PFAM " = '%s';", dims[j]);
  }
  else if (strstr (dims[j], "enog:")) {
    /* set up the query. */
    ret = db_query ("select distinct dsc from " \
                    TBL_NOG " where " F_NOG " = '%s';", dims[j]);
  }
  else if (strstr (dims[j], "reg:")) {
    /* just return a set string. */
    if (strcmp (dims[j], F_REG_UP) == 0) return str_dup ("Up-regulated");
    if (strcmp (dims[j], F_REG_DOWN) == 0) return str_dup ("Down-regulated");
  }

  /* see if the query worked. */
  if (ret) {
    /* print an error message and return nothing. */
    fprintf (stderr, "WARN! unable to query description for '%s'!\n", dims[j]);
    return NULL;
  }

  /* get the query results. */
  struct db_results *results = db_results_get ();

  /* copy the description string out. */
  fval = db_results_field (results, 0, "dsc");
  if (fval)
    desc = str_dup (fval);
  else
    desc = NULL;

  /* free the query results. */
  db_results_free (results);

  /* return the description. */
  return desc;
#else
  /* print a failure message and exit. */
  fprintf (stderr, "FAIL! no mysql connection to describe dimensions!\n");
  exit (1);
#endif
}

/* dim_count: count the members in a given dimension. */
int dim_count (unsigned long j) {
  /* declare required variables. */
  unsigned long i;
  int count;

  /* loop through the members. */
  for (count = 0, i = 0; i < N; i++) if (space[i][j]) count++;

  /* return the count. */
  return count;
}

/* util_init_core: initializes all required structures for calculations. */
void util_init_core (int argc, char **argv) {
  /* check the arguments. */
  if (argc != 2 || !argv[1]) {
    /* fail. */
    fprintf (stderr, "FAIL! i need a conf file.\n");
    exit (1);
  }

  /* print a nice message. */
  fprintf (stderr, "INFO: starting up.\n");

  /* init the rng. */
  rand_init ();

  /* connect to the database. */
#if CONFIGURE_MYSQL == 1
  db_connect (NULL);
#endif

  /* init the config file. */
  conf_init (argv[1]);

  /* check the conf keys. */
  conf_check (CONF_UIDS);
  conf_check (CONF_DIMS);
  conf_check (CONF_SPACE);
  conf_check (CONF_DIST);

  /* init the output filenames. */
  f_uids = conf_get (CONF_UIDS);
  f_dims = conf_get (CONF_DIMS);
  f_space = conf_get (CONF_SPACE);
  f_dist = conf_get (CONF_DIST);

  /* build the uniprot id label array. */
  init_uids ();

#if CONFIGURE_MYSQL == 1
  /* get the known proteins. */
  find_knowns ();
#endif

  /* build the dimensions label array. */
  init_dims ();

  /* build the multidimensional space. */
  init_space ();

  /* disconnect from the database. */
#if CONFIGURE_MYSQL == 1
  db_disconnect ();
#endif

  /* calculate the distances between all proteins. */
  init_dist ();

  /* print a nice message. */
  fprintf (stderr, "INFO: initialization completed.\n");
}

/* read_princomp: reads in the principle component analysis scores space. */
double **read_princomp (unsigned long *n) {
  /* declare required variables. */
  unsigned long i, j, offset;
  double **X = NULL;
  char buf[2048];

  /* print a nice message. */
  fprintf (stderr, "INFO: reading in principle component analysis results.\n");

  /* check the conf key. */
  conf_check (CONF_PRINCOMP);

  /* open the pca output file. */
  FILE *fin = fopen (conf_get (CONF_PRINCOMP), "rb");
  if (!fin) return NULL;

  /* count the number of principle components. */
  *n = 1;
  if (!fgets (buf, 2048, fin)) return NULL;
  for (i = 0; i < strlen (buf); i++) {
    /* increment n on commas. */
    if (buf[i] == ',') (*n)++;
  }

  /* allocate memory for the output matrix. */
  X = (double **) malloc (sizeof (double *) * N);
  for (i = 0; i < N; i++) {
    /* allocate memory for the output row. */
    X[i] = (double *) malloc (sizeof (double) * (*n));
    for (j = 0; j < *n; j++) X[i][j] = 0.0;
  }

  /* rewind the input file. */
  fseek (fin, 0L, SEEK_SET);

  /* read in the file. */
  i = 0;
  while (!feof (fin)) {
    /* read the next line from the file. */
    j = offset = 0;
    if (fgets (buf, 2048, fin)) {
      /* parse the values from the line. */
      while (j < *n) {
        /* get the next value. */
        X[i][j] = atof (buf + offset); j++;

        /* shift to the next delimiter character. */
        while (buf[offset] != ',' && buf[offset] != '\n') offset++;

        /* which delimiter character did we stop on? */
        if (buf[offset] == ',') {
          /* comma, get the next value. */
          offset++;
        }
        else {
          /* newline, increment the line counter. */
          i++;
          break;
        }
      }
    }
  }

  /* close the input file. */
  fclose (fin);

  /* return the collected data. */
  return X;
}

/* read_clustering: reads in a clustering output file. */
unsigned long *read_clustering (const char *fclust) {
  /* declare required variables. */
  unsigned long i, j;
  unsigned long *indices;
  char buf[2048];

  /* open the input file. */
  FILE *fin = fopen (fclust, "rb");
  if (!fin) {
    /* fail. */
    fprintf (stderr, "FAIL! unable to open file '%s'!\n", fclust);
    return NULL;
  }

  /* allocate memory for the indices. */
  indices = (unsigned long *) malloc (sizeof (unsigned long) * N);

  /* loop through the lines of the file. */
  while (!feof (fin)) {
    /* get the line into a buffer string. */
    if (fgets (buf, 2048, fin)) {
      /* see if we're on the first line. */
      if (sscanf (buf, "%lu,%lu,", &i, &j) == 2) indices[i] = j;
    }
  }

  /* close the input file. */
  fclose (fin);

  /* return the output. */
  return indices;
}
